from __future__ import annotations

from enum import Enum
from typing import Annotated, Any, Literal

from pydantic import BaseModel, Field, PrivateAttr, field_validator

from skeletoken.common import RegexPattern, StringPattern, coerce_string_regex_pattern


class NormalizerType(str, Enum):
    SEQUENCE = "Sequence"
    NFC = "NFC"
    NFD = "NFD"
    NFKC = "NFKC"
    NFKD = "NFKD"
    BERTNORMALIZER = "BertNormalizer"
    BYTELEVEL = "ByteLevel"
    LOWERCASE = "Lowercase"
    NMT = "Nmt"
    PREPEND = "Prepend"
    STRIP = "Strip"
    STRIPACCENTS = "StripAccents"
    REPLACE = "Replace"
    PRECOMPILED = "Precompiled"


class BaseNormalizer(BaseModel):
    """Base class for all normalizers."""

    _lowercases: bool = PrivateAttr(default=False)
    _byte_normalizes: bool = PrivateAttr(default=False)

    @property
    def lowercases(self) -> bool:
        """Check if the normalizer lowercases."""
        return self._lowercases

    @property
    def byte_normalizes(self) -> bool:
        """Check if the normalizer byte normalizes."""
        return self._byte_normalizes


class NormalizerSequence(BaseModel):
    """A sequence of normalizers to be applied in order."""

    sequence_field: str = "normalizers"
    type: Literal[NormalizerType.SEQUENCE] = NormalizerType.SEQUENCE
    normalizers: list[NormalizerDiscriminator]

    @property
    def lowercases(self) -> bool:
        """Check if the sequence contains a lowercase normalizer."""
        return any(normalizer.lowercases for normalizer in self.normalizers)

    @property
    def byte_normalizes(self) -> bool:
        """Check if the sequence contains a byte normalizer."""
        return any(normalizer.byte_normalizes for normalizer in self.normalizers)


class NFCNormalizer(BaseNormalizer):
    """
    Applies NFC normalization to the input text.

    See here for more details:
    https://unicode.org/reports/tr15/#Normalization_Forms

    """

    type: Literal[NormalizerType.NFC] = NormalizerType.NFC


class NFDNormalizer(BaseNormalizer):
    """
    Applies NFD normalization to the input text.

    See here for more details:
    https://unicode.org/reports/tr15/#Normalization_Forms

    """

    type: Literal[NormalizerType.NFD] = NormalizerType.NFD


class NFKCNormalizer(BaseNormalizer):
    """
    Applies NFKC normalization to the input text.

    See here for more details:
    https://unicode.org/reports/tr15/#Normalization_Forms

    """

    type: Literal[NormalizerType.NFKC] = NormalizerType.NFKC


class NFKDNormalizer(BaseNormalizer):
    """
    Applies NFKD normalization to the input text.

    See here for more details:
    https://unicode.org/reports/tr15/#Normalization_Forms

    """

    type: Literal[NormalizerType.NFKD] = NormalizerType.NFKD


class BertNormalizer(BaseNormalizer):
    """
    The normalization used by the original BERT.

    Attributes
    ----------
    type : Literal[NormalizerType.BERTNORMALIZER]
        The type of the normalizer. This is always "BertNormalizer".
    clean_text : bool
        If set, this normalizes whitespace and removes any control characters.
    handle_chinese_chars : bool
        If set, it surrounds any Chinese characters in the text with spaces so that they are pretokenized correctly.
    strip_accents : bool | None
        If set, this removes all accents from the text.
    lowercase : bool
        If set, this lowercases the text.

    """

    type: Literal[NormalizerType.BERTNORMALIZER] = NormalizerType.BERTNORMALIZER
    clean_text: bool
    handle_chinese_chars: bool
    strip_accents: bool | None
    lowercase: bool

    @property
    def lowercases(self) -> bool:
        """Check if the normalizer lowercases."""
        return self.lowercase


class ByteLevelNormalizer(BaseNormalizer):
    """
    Applies byte-level normalization to the input text.

    This normalizer applies the same transformations as the ByteLevel pretokenizer.
    Using this normalizer and adding a regex split pretokenizer is equivalent to using the ByteLevel pretokenizer.

    """

    type: Literal[NormalizerType.BYTELEVEL] = NormalizerType.BYTELEVEL
    _byte_normalizes: bool = PrivateAttr(default=True)


class LowercaseNormalizer(BaseNormalizer):
    """Lowercases the input text."""

    type: Literal[NormalizerType.LOWERCASE] = NormalizerType.LOWERCASE
    _lowercases: bool = True


class NmtNormalizer(BaseNormalizer):
    """
    A normalizer that removes specific codepoints.

    The codepoints:
        0x0001..=0x0008 -> Control characters SOH to BS
        0x000B -> Vertical tab
        0x000E..=0x001F -> More control characters
        0x007F -> DEL (delete)
        0x008F, 0x009F -> Control characters from C1 set

    are removed

    The codepoints:
        0x0009 => Tab (Horizontal Tab)
        0x000A => Line Feed (LF / Newline)
        0x000C => Form Feed (FF)
        0x000D => Carriage Return (CR)
        0x1680 => Ogham Space Mark
        0x200B..=0x200F => Zero Width Space and related (ZWSP, ZWNJ, ZWJ, LRM, RLM, etc.)
        0x2028 => Line Separator
        0x2029 => Paragraph Separator
        0x2581 => Lower One Eighth Block (▁) – used as visible space in some tokenizers
        0xFEFF => Zero Width No-Break Space / Byte Order Mark (BOM)
        0xFFFD => Replacement Character (�)

    are replaced with a space character (U+0020).

    """

    type: Literal[NormalizerType.NMT] = NormalizerType.NMT


class PrependNormalizer(BaseNormalizer):
    """
    Prepends a string to the input text.

    Attributes
    ----------
    type : Literal[NormalizerType.PREPEND]
        The type of the normalizer. This is always "Prepend".
    prepend : str
        The string to prepend to a text.

    """

    type: Literal[NormalizerType.PREPEND] = NormalizerType.PREPEND
    prepend: str


class StripNormalizer(BaseNormalizer):
    """
    Strips whitespace from the left and/or right side of the input text.

    Attributes
    ----------
    type : Literal[NormalizerType.STRIP]
        The type of the normalizer. This is always "Strip".
    strip_left : bool
        If set, this removes whitespace from the left side.
    strip_right : bool
        If set, this removes whitespace from the right side.

    """

    type: Literal[NormalizerType.STRIP] = NormalizerType.STRIP
    strip_left: bool
    strip_right: bool


class StripAccentsNormalizer(BaseNormalizer):
    """Strips accents from the input text."""

    type: Literal[NormalizerType.STRIPACCENTS] = NormalizerType.STRIPACCENTS


class ReplaceNormalizer(BaseNormalizer):
    """Replaces a pattern in the input text with a given content."""

    type: Literal[NormalizerType.REPLACE] = NormalizerType.REPLACE
    pattern: StringPattern | RegexPattern
    content: str

    @field_validator("pattern", mode="before")
    @classmethod
    def _coerce_pattern(cls, v: Any) -> StringPattern | RegexPattern | dict:
        return coerce_string_regex_pattern(v)


class PrecompiledNormalizer(BaseNormalizer):
    """
    A precompiled normalizer that uses a precompiled characters map.

    NOTE: It is unclear how this is constructed, and is mainly here for compatibility with sentencepiece

    Attributes
    ----------
    type : Literal[NormalizerType.PRECOMPILED]
        The type of the normalizer. This is always "Precompiled".
    precompiled_charsmap : str
        The precompiled characters map.

    """

    type: Literal[NormalizerType.PRECOMPILED] = NormalizerType.PRECOMPILED
    precompiled_charsmap: str


Normalizer = (
    NFCNormalizer
    | NFDNormalizer
    | NFKCNormalizer
    | NFKDNormalizer
    | BertNormalizer
    | ByteLevelNormalizer
    | LowercaseNormalizer
    | NmtNormalizer
    | PrependNormalizer
    | StripNormalizer
    | StripAccentsNormalizer
    | ReplaceNormalizer
    | PrecompiledNormalizer
    | NormalizerSequence
)
NormalizerDiscriminator = Annotated[Normalizer, Field(discriminator="type")]



import logging
from typing import Any

from pydantic import PrivateAttr, RootModel

logger = logging.getLogger(__name__)

_Merge = tuple[str, str]


class Merges(RootModel[list[_Merge]]):
    """Data model representing a list of merge operations."""

    _merge_index: dict[_Merge, int] = PrivateAttr(default_factory=dict)
    _all_merge_tokens: set[str] = PrivateAttr(default_factory=set)

    def model_post_init(self, context: Any) -> None:
        """Post init of the model."""
        self._merge_index = {merge: i for i, merge in enumerate(self.root)}
        self._all_merge_tokens = set()
        for left, right in self.root:
            self._all_merge_tokens.add(left)
            self._all_merge_tokens.add(right)

    def _add_merges_for_token(self, token: str) -> list[str]:
        """Add merge operations for a specific token."""
        # Convert to tuple for typing.
        token_form = tuple(self._merge(token))
        added_merges = []
        while len(token_form) > 1:
            for index in range(0, len(token_form) - 1, 2):
                left, right = token_form[index], token_form[index + 1]
                bigram = (left, right)
                self.root.append(bigram)
                self._merge_index[bigram] = len(self.root) - 1
                if left not in self._all_merge_tokens:
                    added_merges.append(left)
                if right not in self._all_merge_tokens:
                    added_merges.append(right)
                self._all_merge_tokens.add(left)
                self._all_merge_tokens.add(right)
            token_form = tuple(self._merge(token))
        return sorted(set(added_merges))

    def _merge(self, token: str) -> list[str]:
        """Merge a token with its subword components."""
        chars = list(token)

        # Special case, the token is length 1.
        if len(chars) == 1:
            return chars

        while True:
            string_bigrams = _bigrams(chars)
            lowest_idx = float("inf")
            merge: _Merge = ("", "")
            for bigram in string_bigrams:
                if bigram not in self._merge_index:
                    continue
                idx = self._merge_index[bigram]
                if idx < lowest_idx:
                    lowest_idx = idx
                    merge = bigram
            if lowest_idx == float("inf"):
                break
            chars = _merge(chars, merge)

        return chars

    def _remove_merges_for_token(self, token: str) -> None:
        """Remove merges from the list of merges."""
        # If the token is part of the merge index, other tokens depend on it.
        # This makes the model crash.
        if token in self._all_merge_tokens:
            new_root = []
            for merge in self.root:
                joined_merge = "".join(merge)
                if token not in merge and token != joined_merge:
                    new_root.append(merge)

            self.root = new_root
            # Rebuild the index.
            self._merge_index = {merge: i for i, merge in enumerate(self.root)}
            self._all_merge_tokens.remove(token)


def _bigrams(chars: list[str]) -> list[_Merge]:
    """Calculate the bigrams of a string."""
    return list(zip(chars[:-1], chars[1:], strict=True))


def _merge(token: list[str], bigram: _Merge) -> list[str]:
    """Merge a bytegram into a token."""
    # Merge the bytegram into the token
    new_token = []
    index = 0
    while index < len(token) - 1:
        if (token[index], token[index + 1]) == bigram:
            new_token.append("".join(bigram))
            index += 1
        else:
            new_token.append(token[index])
        index += 1
    if index < len(token):
        new_token.append(token[index])
    return new_token



import pytest

from skeletoken.merges import Merges


def test_merges(small_merges: Merges) -> None:
    """Test simple merge functionality."""
    assert small_merges.root == [("a", "b"), ("c", "d"), ("ab", "c"), ("a", "bc")]
    assert small_merges._merge_index == {
        ("a", "b"): 0,
        ("c", "d"): 1,
        ("ab", "c"): 2,
        ("a", "bc"): 3,
    }


def test_merge(small_merges: Merges) -> None:
    """Test the merge function."""
    assert small_merges._merge("abc") == ["abc"]
    # Single character just returns
    assert small_merges._merge("a") == ["a"]
    assert small_merges._merge("abcd") == ["ab", "cd"]
    assert small_merges._merge("bca") == ["b", "c", "a"]
    # Nothing in vocab just returns all single characters
    assert small_merges._merge("xyz") == ["x", "y", "z"]


@pytest.mark.parametrize(
    "token, before, added",
    [
        ("ab", ["ab"], [("a", "b"), ("c", "d"), ("ab", "c"), ("a", "bc")]),
        ("cd", ["cd"], [("a", "b"), ("c", "d"), ("ab", "c"), ("a", "bc")]),
        ("abc", ["abc"], [("a", "b"), ("c", "d"), ("ab", "c"), ("a", "bc")]),
        ("bc", ["b", "c"], [("a", "b"), ("c", "d"), ("ab", "c"), ("a", "bc"), ("b", "c")]),
        (
            "elephantine",
            ["e", "l", "e", "p", "h", "a", "n", "t", "i", "n", "e"],
            [
                ("a", "b"),
                ("c", "d"),
                ("ab", "c"),
                ("a", "bc"),
                ("e", "l"),
                ("e", "p"),
                ("h", "a"),
                ("n", "t"),
                ("i", "n"),
                ("el", "ep"),
                ("ha", "nt"),
                ("in", "e"),
                ("elep", "hant"),
                ("elephant", "ine"),
            ],
        ),
    ],
)
def test_add_merges(small_merges: Merges, token: str, before: list[str], added: list[str]) -> None:
    """Test whether merges are added."""
    assert small_merges._merge(token) == before
    small_merges._add_merges_for_token(token)
    assert small_merges._merge(token) == [token]
    assert small_merges.root == added


def test_remove_merges(small_merges: Merges) -> None:
    """Test whether merges are removed."""
    # Original merges are [("a", "b"), ("c", "d"), ("ab", "c"), ("a", "bc")]
    # For the string "ab", we remove ("a", "b") and ("ab", "c")
    # We remove ("a", "b") because "ab" is the result of that merge
    # We also remove ("ab", "c") because it contains "ab" as a subpart
    small_merges._remove_merges_for_token("ab")
    assert small_merges.root == [("c", "d"), ("a", "bc")]



import json
from pathlib import Path
from typing import Any

from pytest import FixtureRequest, fixture
from tokenizers import Tokenizer
from transformers import PreTrainedTokenizerFast

from skeletoken.merges import Merges


def _get_path(name: str) -> Path:
    """Gets the path to a test resource."""
    return Path("tests") / "data" / name


def _get_tokenizers_path(name: str) -> str:
    """Gets the path to a test tokenizer resource."""
    return str(_get_path(name) / "tokenizer.json")


@fixture(scope="function")
def small_tokenizer_json(request: FixtureRequest) -> dict[str, Any]:
    """Load an extremely small tokenizer for testing purposes."""
    name = getattr(request, "param", None) or "wordpiece"
    return json.load(open(_get_tokenizers_path(name), "r", encoding="utf-8"))


@fixture(scope="function")
def small_tokenizer(request: FixtureRequest) -> Tokenizer:
    """Load an extremely small tokenizer for testing purposes."""
    name = getattr(request, "param", None) or "wordpiece"
    return Tokenizer.from_file(_get_tokenizers_path(name))


@fixture(scope="function")
def small_merges() -> Merges:
    """Load an extremely small set of merges for testing purposes."""
    return Merges([("a", "b"), ("c", "d"), ("ab", "c"), ("a", "bc")])


@fixture(scope="function")
def transformers_tokenizer(request: FixtureRequest) -> PreTrainedTokenizerFast:
    """Load a small transformers tokenizer for testing purposes."""
    name = getattr(request, "param", None) or "wordpiece"
    tokenizer = PreTrainedTokenizerFast.from_pretrained(_get_path(name))
    return tokenizer



import pytest

from skeletoken.vocabulary import UnigramVocabulary, Vocabulary


def _simple_vocabulary_fixture() -> Vocabulary:
    """Simple pseudofixture to be used in a parametrize."""
    return Vocabulary({"hello": 0, "world": 1})


def _simple_unigram_fixture() -> UnigramVocabulary:
    """Simple pseudofixture to be used in a parametrize."""
    return UnigramVocabulary([("hello", -0.1), ("world", -0.2)])


@pytest.mark.parametrize("vocab", [_simple_vocabulary_fixture(), _simple_unigram_fixture()])
def test_vocabulary(vocab: Vocabulary | UnigramVocabulary) -> None:
    """Test the vocabulary implementation."""
    assert vocab.vocabulary == {"hello": 0, "world": 1}
    assert vocab.sorted_vocabulary == ["hello", "world"]


@pytest.mark.parametrize("vocab", [_simple_vocabulary_fixture(), _simple_unigram_fixture()])
def test_add_token(vocab: Vocabulary | UnigramVocabulary) -> None:
    """Test the vocabulary implementation."""
    vocab.add_token("new_token")
    assert vocab.vocabulary["new_token"] == len(vocab.vocabulary) - 1
    assert vocab.sorted_vocabulary == ["hello", "world", "new_token"]
    with pytest.raises(ValueError):
        vocab.add_token("new_token")


@pytest.mark.parametrize("vocab", [_simple_vocabulary_fixture(), _simple_unigram_fixture()])
def test_remove_token(vocab: Vocabulary | UnigramVocabulary) -> None:
    """Test the vocabulary implementation."""
    vocab.add_token("new_token")
    assert vocab.vocabulary["new_token"] == len(vocab.vocabulary) - 1
    assert vocab.sorted_vocabulary == ["hello", "world", "new_token"]
    vocab.remove_token("new_token")
    assert "new_token" not in vocab.vocabulary
    assert vocab.sorted_vocabulary == ["hello", "world"]
    with pytest.raises(ValueError):
        vocab.remove_token("new_token")


@pytest.mark.parametrize("vocab", [_simple_vocabulary_fixture(), _simple_unigram_fixture()])
def test_replace_token(vocab: Vocabulary | UnigramVocabulary) -> None:
    """Test the vocabulary implementation."""
    vocab.add_token("new_token")
    assert vocab.vocabulary["new_token"] == len(vocab.vocabulary) - 1
    assert vocab.sorted_vocabulary == ["hello", "world", "new_token"]
    vocab.replace_token("new_token", "another_token")
    assert "new_token" not in vocab.vocabulary
    assert vocab.vocabulary["another_token"] == len(vocab.vocabulary) - 1
    assert vocab.sorted_vocabulary == ["hello", "world", "another_token"]
    with pytest.raises(ValueError):
        vocab.replace_token("new_token", "another_token")
    with pytest.raises(ValueError):
        vocab.replace_token("hello", "hello")


@pytest.mark.parametrize("vocab", [_simple_vocabulary_fixture(), _simple_unigram_fixture()])
def test_replace_vocabulary(vocab: Vocabulary | UnigramVocabulary) -> None:
    """Test the vocabulary implementation."""
    assert vocab.sorted_vocabulary == ["hello", "world"]
    vocab.replace_vocabulary(["new_token", "hello"])
    assert vocab.sorted_vocabulary == ["new_token", "hello"]


@pytest.mark.parametrize("vocab", [_simple_vocabulary_fixture(), _simple_unigram_fixture()])
def test_in(vocab: Vocabulary | UnigramVocabulary) -> None:
    """Test the vocabulary implementation."""
    assert "hello" in vocab
    assert "world" in vocab
    assert "new_token" not in vocab


@pytest.mark.parametrize("vocab", [_simple_vocabulary_fixture(), _simple_unigram_fixture()])
def test_getitem(vocab: Vocabulary | UnigramVocabulary) -> None:
    """Test the vocabulary implementation."""
    assert vocab["hello"] == 0
    assert vocab["world"] == 1
    with pytest.raises(KeyError):
        vocab["new_token"]


@pytest.mark.parametrize("vocab", [_simple_vocabulary_fixture(), _simple_unigram_fixture()])
def test_len(vocab: Vocabulary | UnigramVocabulary) -> None:
    """Test the vocabulary implementation."""
    assert len(vocab) == 2



from typing import Any

from skeletoken.addedtoken import AddedToken, AddedTokens
from skeletoken.base import TokenizerModel
from skeletoken.models import WordPiece


def test_addedtoken(small_tokenizer_json: dict[str, Any]) -> None:
    """
    Test that the small tokenizer JSON can be loaded and contains the expected structure.

    This test checks that the tokenizer JSON has the correct keys and types for its fields.
    """
    small_tokenizer_json["added_tokens"] = []
    token_a = AddedToken(
        content="a", single_word=True, lstrip=False, rstrip=False, normalized=False, special=False, id=11
    )
    token_b = AddedToken(
        content="b", single_word=True, lstrip=False, rstrip=False, normalized=False, special=False, id=10
    )
    toks = [token_a, token_b]
    for token in toks:
        small_tokenizer_json["added_tokens"].append(token.model_dump())

    tokenizer = TokenizerModel.model_validate(small_tokenizer_json)
    assert isinstance(tokenizer.model, WordPiece)
    tokenizer.model.continuing_subword_prefix = ""
    assert len(tokenizer.added_tokens) == 3
    for a, b in zip(tokenizer.added_tokens.root[:-1], toks, strict=True):
        assert a.content == b.content
        assert a.single_word == b.single_word
        assert a.rstrip == b.rstrip
        assert a.lstrip == b.lstrip
        assert a.normalized == b.normalized
        assert a.special == b.special
        assert a.id == tokenizer.model.vocab[b.content]

    # Implicit test. If this fails, the model is incorrect.
    tok = tokenizer.to_tokenizer()

    assert tok.encode("a b c").tokens == ["a", " ", "b", " ", "c"]


def test_addedtokens_object() -> None:
    """Test the AddedTokens object."""
    tokens = AddedTokens([])
    tokens.maybe_add_token("a", 11)
    assert len(tokens) == 1
    assert tokens.root[0].id == 11
    tokens.maybe_add_token("a", 12)
    assert len(tokens) == 1
    assert tokens.root[0].id == 12

    tokens.maybe_remove_token("a")
    assert len(tokens) == 0
    assert tokens.root == []
    # Should not raise an error
    tokens.maybe_remove_token("a")
    tokens.maybe_replace_token("a", "b")

    tokens.maybe_add_token("a", 10)
    tokens.maybe_replace_token("a", "b")
    assert len(tokens) == 1
    assert tokens.root[0].content == "b"

    tokens.maybe_add_token("b", 10, normalized=True)
    assert len(tokens) == 1
    assert tokens.root[0].id == 10
    assert tokens.root[0].normalized is True

    assert tokens[0] == tokens.root[0]



from skeletoken.decase.byte_handlers import text_to_token_str, token_to_bytes


def _determine_collision(
    token: str, is_byte: bool, vocab: set[str], special_tokens: list[str], seen: set[str], remove_collisions: bool
) -> str | None:
    """Determine whether a given token, when lowered, collides with another."""
    if token in special_tokens:
        return token
    if is_byte:
        # Convert token from bytes to a string.
        try:
            token_bytes = token_to_bytes(token)
        except ValueError:
            return token
        try:
            new_token = token_bytes.decode("utf-8")
        except UnicodeDecodeError:
            return token
    else:
        new_token = token

    lowered_token = new_token.casefold()

    if is_byte:
        lowered_token = text_to_token_str(lowered_token)

    # If the token changed but the lowered version was already in vocab, we have a collision.
    if (lowered_token != token) and (lowered_token in vocab or lowered_token in seen):
        return None if remove_collisions else token
    return lowered_token


def decase_vocabulary(
    vocabulary: list[str], special_tokens: list[str], is_byte: bool, remove_collisions: bool = False
) -> list[str]:
    """Lowercase the vocabulary of a tokenizer."""
    uncased_vocab: list[str] = []
    # seen keeps track of lowered tokens that were not in vocab before.
    # e.g., "AB" and "Ab" are in vocab, but they collide after lowercasing
    seen: set[str] = set()
    vocab_set = set(vocabulary)

    for token in vocabulary:
        lowered = _determine_collision(token, is_byte, vocab_set, special_tokens, seen, remove_collisions)
        if lowered is None:
            continue
        uncased_vocab.append(lowered)
        seen.add(lowered)

    return uncased_vocab



from pydantic import BaseModel

from skeletoken.base import TokenizerModel


class ModelDelta(BaseModel):
    """
    Represents the differences between two tokenizer models.

    This is used to remap indices in the embedding table when loading.

    Attributes
    ----------
    token_mapping : dict[int, int]
        A mapping from old token IDs to new token IDs.
    new_tokens : dict[str, int]
        A mapping from new token strings to their IDs. This is useful for inferring
        contents of the new tokens.
    new_vocabulary_size : int
        The size of the new vocabulary in the modified model.

    """

    # Used to remap from old token ids to new token ids.
    token_mapping: dict[int, int]
    # Used to identify which tokens go to which new ids.
    new_tokens: dict[str, int]
    # The new vocabulary size. This is used to resize the embedding table.
    new_vocabulary_size: int


def compute_model_delta(original: TokenizerModel, modified: TokenizerModel) -> ModelDelta:
    """Compute the delta between two tokenizer models."""
    # Compute the mapping
    token_mapping = {}
    new_tokens = {}
    new_vocab = modified.model.vocab.vocabulary
    old_vocab = original.model.vocab.vocabulary
    for token, new_id in new_vocab.items():
        if token in old_vocab:
            old_id = old_vocab[token]
            if old_id != new_id:
                token_mapping[old_id] = new_id
        else:
            new_tokens[token] = new_id
    # Compute new tokens
    new_vocabulary_size = len(modified.model.vocab)

    return ModelDelta(token_mapping=token_mapping, new_tokens=new_tokens, new_vocabulary_size=new_vocabulary_size)



"""Utilities to handle byte conversion in decasing."""


def bytes_to_unicode() -> dict[int, str]:
    """Converts byte values to Unicode characters for byte-level tokenization."""
    input_ids = list(range(33, 127)) + list(range(161, 173)) + list(range(174, 256))
    output_ids = input_ids[:]
    n = 0
    for char_ord in range(256):
        if char_ord not in input_ids:
            input_ids.append(char_ord)
            output_ids.append(256 + n)
            n += 1
    output_chars = [chr(c) for c in output_ids]
    return dict(zip(input_ids, output_chars, strict=True))


_MAPPING = bytes_to_unicode()
_INV_MAPPING = {v: k for k, v in _MAPPING.items()}


def token_to_bytes(token: str) -> bytes:
    """Convert a token string to bytes using the byte-level mapping."""
    # Each character in a byte-level token maps to exactly one byte.
    # We need to use .get because some special characters may not be in the mapping.
    # This will return 0 for any character not in the mapping.
    try:
        return bytes(_INV_MAPPING[char] for char in token)
    except KeyError as e:
        raise ValueError(f"Token contains characters that cannot be mapped to bytes: {token}") from e


def text_to_token_str(text: str) -> str:
    """Convert a text string to a token string using the byte-level mapping."""
    # Encode text as UTF-8 bytes, then map each byte back to the placeholder char.
    encoded = text.encode("utf-8")
    return "".join(_MAPPING[byte] for byte in encoded)



from pydantic import PrivateAttr, RootModel


class VocabMixin:
    """Mixin class for vocabulary-related functionality."""

    @property
    def vocabulary(self) -> dict[str, int]:
        """Returns the vocabulary mapping."""
        raise NotImplementedError()  # pragma: no cover

    def __contains__(self, token: str) -> bool:
        """Checks if a token is in the vocabulary."""
        return token in self.vocabulary

    def __getitem__(self, token: str) -> int:
        """Gets the ID of a token in the vocabulary."""
        return self.vocabulary[token]

    def __len__(self) -> int:
        """Gets the number of tokens in the vocabulary."""
        return len(self.vocabulary)


class Vocabulary(RootModel[dict[str, int]], VocabMixin):
    """A vocabulary mapping tokens to their IDs."""

    @property
    def vocabulary(self) -> dict[str, int]:
        """Returns the vocabulary mapping."""
        return self.root

    @property
    def sorted_vocabulary(self) -> list[str]:
        """Returns the vocabulary mapping sorted by token."""
        return [x[0] for x in sorted(self.root.items(), key=lambda x: x[1])]

    def add_token(self, token: str) -> None:
        """Adds a token to the vocabulary."""
        if token in self.root:
            raise ValueError(f"Token '{token}' already exists in vocabulary.")
        self.root[token] = len(self.root)

    def replace_token(self, old_token: str, new_token: str) -> None:
        """Replaces tokens."""
        if old_token not in self.root:
            raise ValueError(f"Token '{old_token}' does not exist in vocabulary.")
        if new_token in self.root:
            raise ValueError(f"Token '{new_token}' already exists in vocabulary.")
        idx = self.root.pop(old_token)
        self.root[new_token] = idx

    def remove_token(self, token: str) -> None:
        """Removes tokens from the vocabulary."""
        if token not in self.root:
            raise ValueError(f"Token '{token}' does not exist in vocabulary.")
        self.root.pop(token)
        # Rebuild the vocabulary to ensure indices are contiguous
        sorted_tokens, _ = zip(*sorted(self.root.items(), key=lambda x: x[1]), strict=True)
        self.root = {token: idx for idx, token in enumerate(sorted_tokens)}

    def replace_vocabulary(self, vocabulary: list[str]) -> None:
        """Completely replaces the vocabulary by a vocabulary of the same length."""
        self.root = {token: idx for idx, token in enumerate(vocabulary)}


class UnigramVocabulary(RootModel[list[tuple[str, float]]], VocabMixin):
    """A unigram vocabulary mapping tokens to their scores."""

    _vocabulary: dict[str, int] = PrivateAttr(default_factory=dict, init=False)
    _min_score: float = PrivateAttr(default=-100.0, init=False)

    @property
    def vocabulary(self) -> dict[str, int]:
        """Returns the vocabulary mapping."""
        return self._vocabulary

    @property
    def sorted_vocabulary(self) -> list[str]:
        """Returns the vocabulary mapping sorted by token."""
        return [x[0] for x in sorted(self._vocabulary.items(), key=lambda x: x[1])]

    def model_post_init(self, __context: dict) -> None:
        """Initializes the vocabulary."""
        tokens, scores = zip(*self.root, strict=True) if self.root else ([], [])
        self._vocabulary = {token: idx for idx, token in enumerate(tokens)}
        self._min_score = min(scores, default=-100.0)

    def add_token(self, token: str) -> None:
        """Adds a token to the vocabulary."""
        if token in self._vocabulary:
            raise ValueError(f"Token '{token}' already exists in vocabulary.")
        self.root.append((token, self._min_score))
        self._vocabulary[token] = len(self.root) - 1

    def replace_token(self, old_token: str, new_token: str) -> None:
        """Removes a token from the vocabulary."""
        if old_token not in self._vocabulary:
            raise ValueError(f"Token '{old_token}' does not exist in vocabulary.")
        if new_token in self._vocabulary:
            raise ValueError(f"Token '{new_token}' already exists in vocabulary.")
        idx = self._vocabulary.pop(old_token)
        self._vocabulary[new_token] = idx
        self.root[idx] = (new_token, self.root[idx][1])

    def remove_token(self, token: str) -> None:
        """Removes a token from the vocabulary."""
        if token not in self._vocabulary:
            raise ValueError(f"Token '{token}' does not exist in vocabulary.")
        idx = self._vocabulary.pop(token)
        sorted_tokens, _ = zip(*sorted(self._vocabulary.items(), key=lambda x: x[1]), strict=True)
        self._vocabulary = {token: idx for idx, token in enumerate(sorted_tokens)}

        self.root.pop(idx)

    def replace_vocabulary(self, vocabulary: list[str]) -> None:
        """Completely replaces the vocabulary by a vocabulary of the same length."""
        self.root = [(token, self.root[idx][1]) for idx, token in enumerate(vocabulary)]
        self._vocabulary = {token: idx for idx, token in enumerate(vocabulary)}



import re

import pytest
import regex

from skeletoken.common import coerce_string_regex_pattern


def test_coercion_to_string_regex() -> None:
    """Tests the coercion function for string and regex patterns."""
    assert coerce_string_regex_pattern("test") == {"String": "test"}
    assert coerce_string_regex_pattern(re.compile("test")) == {"Regex": "test"}
    assert coerce_string_regex_pattern(regex.compile("test")) == {"Regex": "test"}

    with pytest.raises(TypeError):
        assert coerce_string_regex_pattern(123)  # type: ignore  # On purpose



from typing import Any

import pytest

from skeletoken.base import TokenizerModel
from skeletoken.common import PrependScheme, StringPattern
from skeletoken.decoders import (
    BPEDecoder,
    ByteFallbackDecoder,
    ByteLevelDecoder,
    CTCDecoder,
    Decoder,
    DecoderType,
    FuseDecoder,
    MetaspaceDecoder,
    ReplaceDecoder,
    StripDecoder,
    WordPieceDecoder,
)


def _get_default_decoder(decoder_type: DecoderType) -> Decoder:
    """Helper function to get the default instantiation of a decoder."""
    if decoder_type == DecoderType.BPEDECODER:
        return BPEDecoder(suffix=r"\w")
    elif decoder_type == DecoderType.BYTEFALLBACK:
        return ByteFallbackDecoder()
    elif decoder_type == DecoderType.BYTELEVEL:
        return ByteLevelDecoder(add_prefix_space=False, trim_offsets=False, use_regex=False)
    elif decoder_type == DecoderType.CTC:
        return CTCDecoder(pad_token="[PAD]", word_delimiter_token="", cleanup=False)
    elif decoder_type == DecoderType.FUSE:
        return FuseDecoder()
    elif decoder_type == DecoderType.METASPACE:
        return MetaspaceDecoder(replacement=" ", prepend_scheme=PrependScheme.FIRST, split=False)
    elif decoder_type == DecoderType.REPLACE:
        return ReplaceDecoder(pattern=StringPattern(String="a"), content="replacement")
    elif decoder_type == DecoderType.STRIP:
        return StripDecoder(content=" ", start=0, stop=1)
    elif decoder_type == DecoderType.WORDPIECE:
        return WordPieceDecoder(prefix="##", cleanup=True)
    else:
        raise ValueError(f"Unknown decoder type: {decoder_type}")


@pytest.mark.parametrize(
    "decoder_type",
    [
        DecoderType.BPEDECODER,
        DecoderType.BYTEFALLBACK,
        DecoderType.BYTELEVEL,
        DecoderType.CTC,
        DecoderType.FUSE,
        DecoderType.METASPACE,
        DecoderType.REPLACE,
        DecoderType.STRIP,
        DecoderType.WORDPIECE,
    ],
)
def test_decoder(small_tokenizer_json: dict[str, Any], decoder_type: DecoderType) -> None:
    """
    Test that the small tokenizer JSON can be loaded and contains the expected structure.

    This test checks that the tokenizer JSON has the correct keys and types for its fields.
    """
    decoder = _get_default_decoder(decoder_type)
    decoder_dict = decoder.model_dump()
    small_tokenizer_json["decoder"] = decoder_dict
    model = TokenizerModel.model_validate(small_tokenizer_json)

    assert model.decoder is not None
    assert model.decoder.type == decoder_type

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()



from enum import Enum
from typing import Literal

from pydantic import BaseModel, RootModel


class BatchLongestStrategy(RootModel[Literal["BatchLongest"]]):
    """Pad to the longest string in the sequence."""

    ...


class FixedStrategy(BaseModel):
    """Pad to a fixed amount."""

    Fixed: int


# Union of the two shapes
PaddingStrategy = BatchLongestStrategy | FixedStrategy


class PaddingDirection(str, Enum):
    LEFT = "Left"
    RIGHT = "Right"


class Padding(BaseModel):
    """
    The padding configuration.

    Attributes
    ----------
    strategy : PaddingStrategy
        The padding strategy to use.
        If this is BatchLongest, the input will be padded to the longest sequence in the batch.
        If this is Fixed, the padding will be applied to a fixed amount. This allows for a
        clever hack. By setting the strategy to Fixed, and the amount to 0, you can set a padding
        token without actually padding.
    direction : PaddingDirection
        The direction to pad in. This can be either left or right.
    pad_to_multiple_of : int | None
        If set, the input will be padded to a multiple of this value.
    pad_id : int
        The ID of the padding token in the vocabulary.
    pad_type_id : int
        The type ID to insert when padding. This is usually 0.
    pad_token : str
        The form of the padding token in the vocabulary.

    """

    strategy: PaddingStrategy = FixedStrategy(Fixed=0)
    direction: PaddingDirection = PaddingDirection.RIGHT
    pad_to_multiple_of: int | None = None
    pad_id: int
    pad_type_id: int
    pad_token: str



from typing import Any

import pytest

from skeletoken.base import TokenizerModel
from skeletoken.common import PrependScheme, StringPattern
from skeletoken.pretokenizers import (
    Behavior,
    BertPreTokenizer,
    ByteLevelPreTokenizer,
    CharDelimiterSplitPreTokenizer,
    DigitsPreTokenizer,
    FixedLengthPreTokenizer,
    MetaspacePreTokenizer,
    PreTokenizer,
    PreTokenizerSequence,
    PreTokenizerType,
    PunctuationPreTokenizer,
    SplitPreTokenizer,
    UnicodeScriptsPreTokenizer,
    WhitespacePreTokenizer,
    WhitespaceSplitPreTokenizer,
    get_metaspace,
)


def _get_default_pretokenizer(pretokenizer_type: PreTokenizerType) -> PreTokenizer:  # noqa: C901
    """Helper function to get the default instantiation of a pretokenizer."""
    if pretokenizer_type == PreTokenizerType.BERT_PRETOKENIZER:
        return BertPreTokenizer()
    elif pretokenizer_type == PreTokenizerType.BYTELEVEL:
        return ByteLevelPreTokenizer(add_prefix_space=False, use_regex=False, trim_offsets=False)
    elif pretokenizer_type == PreTokenizerType.CHARDELIMITERSPLIT:
        return CharDelimiterSplitPreTokenizer(delimiter="|")
    elif pretokenizer_type == PreTokenizerType.DIGITS:
        return DigitsPreTokenizer(individual_digits=True)
    elif pretokenizer_type == PreTokenizerType.FIXEDLENGTH:
        return FixedLengthPreTokenizer(length=5)
    elif pretokenizer_type == PreTokenizerType.METASPACE:
        return MetaspacePreTokenizer(replacement=" ", prepend_scheme=PrependScheme.FIRST, split=True)
    elif pretokenizer_type == PreTokenizerType.PUNCTUATION:
        return PunctuationPreTokenizer(behavior=Behavior.CONTIGUOUS)
    elif pretokenizer_type == PreTokenizerType.SPLIT:
        return SplitPreTokenizer(pattern=StringPattern(String="a"), behavior=Behavior.ISOLATED, invert=False)
    elif pretokenizer_type == PreTokenizerType.WHITESPACE:
        return WhitespacePreTokenizer()
    elif pretokenizer_type == PreTokenizerType.WHITESPACESPLIT:
        return WhitespaceSplitPreTokenizer()
    elif pretokenizer_type == PreTokenizerType.UNICODESCRIPTS:
        return UnicodeScriptsPreTokenizer()
    else:
        raise ValueError(f"Unknown pretokenizer type: {pretokenizer_type}")


@pytest.mark.parametrize(
    "pretokenizer_type",
    [
        PreTokenizerType.BYTELEVEL,
        PreTokenizerType.BERT_PRETOKENIZER,
        PreTokenizerType.CHARDELIMITERSPLIT,
        PreTokenizerType.DIGITS,
        PreTokenizerType.FIXEDLENGTH,
        PreTokenizerType.METASPACE,
        PreTokenizerType.PUNCTUATION,
        PreTokenizerType.SPLIT,
        PreTokenizerType.WHITESPACE,
        PreTokenizerType.WHITESPACESPLIT,
        PreTokenizerType.UNICODESCRIPTS,
    ],
)
def test_pretokenizer(small_tokenizer_json: dict[str, Any], pretokenizer_type: PreTokenizerType) -> None:
    """
    Test that the small tokenizer JSON can be loaded and contains the expected structure.

    This test checks that the tokenizer JSON has the correct keys and types for its fields.
    """
    normalizer = _get_default_pretokenizer(pretokenizer_type)
    normalizer_dict = normalizer.model_dump()
    small_tokenizer_json["pre_tokenizer"] = normalizer_dict
    tokenizer = TokenizerModel.model_validate(small_tokenizer_json)

    assert tokenizer.pre_tokenizer is not None
    assert tokenizer.pre_tokenizer.type == pretokenizer_type

    # Implicit test. If this fails, the model is incorrect.
    tokenizer.to_tokenizer()


@pytest.mark.parametrize(
    "pretokenizer,should_byte_transform",
    [
        [_get_default_pretokenizer(PreTokenizerType.BERT_PRETOKENIZER), False],
        [_get_default_pretokenizer(PreTokenizerType.BYTELEVEL), True],
        [PreTokenizerSequence(pretokenizers=[_get_default_pretokenizer(PreTokenizerType.BYTELEVEL)]), True],
        [
            PreTokenizerSequence(
                pretokenizers=[
                    PreTokenizerSequence(pretokenizers=[_get_default_pretokenizer(PreTokenizerType.BYTELEVEL)])
                ]
            ),
            True,
        ],
    ],
)
def test_byte_transform(pretokenizer: PreTokenizer, should_byte_transform: bool) -> None:
    """Test whether the byte transform detection works."""
    assert pretokenizer._byte_pretokenizes == should_byte_transform


@pytest.mark.parametrize(
    "pretokenizer,splits",
    [
        [_get_default_pretokenizer(PreTokenizerType.BERT_PRETOKENIZER), True],
        [_get_default_pretokenizer(PreTokenizerType.BYTELEVEL), False],
        [PreTokenizerSequence(pretokenizers=[_get_default_pretokenizer(PreTokenizerType.BYTELEVEL)]), False],
        [
            PreTokenizerSequence(
                pretokenizers=[
                    PreTokenizerSequence(pretokenizers=[_get_default_pretokenizer(PreTokenizerType.BYTELEVEL)])
                ]
            ),
            False,
        ],
        [_get_default_pretokenizer(PreTokenizerType.METASPACE), True],
    ],
)
def test_splits(pretokenizer: PreTokenizer, splits: bool) -> None:
    """Test whether a pretokenizer splits."""
    assert pretokenizer._splits == splits


def test_get_metaspace() -> None:
    """Test detection of a metaspace."""
    default = _get_default_pretokenizer(PreTokenizerType.METASPACE)
    assert get_metaspace(default) == " "
    assert get_metaspace(PreTokenizerSequence(pretokenizers=[default])) == " "

    other_default = _get_default_pretokenizer(PreTokenizerType.BERT_PRETOKENIZER)
    assert get_metaspace(other_default) is None
    assert get_metaspace(PreTokenizerSequence(pretokenizers=[other_default])) is None



from skeletoken.decase.decase import _determine_collision, decase_vocabulary


def test_determine_collision() -> None:
    """Test the collision detection in the vocabulary."""
    vocabulary = {"dog", "CAT", "Cat", "cat"}
    # Unchanged (already in vocab)
    assert (
        _determine_collision(
            token="dog", is_byte=False, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "dog"
    )
    # Unchanged (already in vocab)
    assert (
        _determine_collision(
            token="DOG", is_byte=False, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "DOG"
    )
    # Unchanged (already lowered)
    assert (
        _determine_collision(
            token="cat", is_byte=False, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "cat"
    )
    # Unchanged (already in vocab)
    assert (
        _determine_collision(
            token="CAT", is_byte=False, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "CAT"
    )
    # Unchanged (already in vocab)
    assert (
        _determine_collision(
            token="cAT", is_byte=False, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "cAT"
    )
    # Additional test (new token)
    assert (
        _determine_collision(
            token="SPIN", is_byte=False, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "spin"
    )
    # Unchanged (already in seen)
    assert (
        _determine_collision(
            token="SPIN",
            is_byte=False,
            vocab=set(vocabulary),
            special_tokens=[],
            seen={"spin"},
            remove_collisions=False,
        )
        == "SPIN"
    )
    # Unchanged (special token)
    assert (
        _determine_collision(
            token="Spin",
            is_byte=False,
            vocab=set(vocabulary),
            special_tokens=["Spin"],
            seen=set(),
            remove_collisions=False,
        )
        == "Spin"
    )

    # Byte stuff
    assert (
        _determine_collision(
            token="dog", is_byte=True, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "dog"
    )
    assert (
        _determine_collision(
            token="DOG", is_byte=True, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "DOG"
    )
    assert (
        _determine_collision(
            token="cat", is_byte=True, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "cat"
    )
    assert (
        _determine_collision(
            token="CAT", is_byte=True, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "CAT"
    )
    assert (
        _determine_collision(
            token="cAT", is_byte=True, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "cAT"
    )
    assert (
        _determine_collision(
            token="SPIN", is_byte=True, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "spin"
    )
    assert (
        _determine_collision(
            token="SPIN", is_byte=True, vocab=set(vocabulary), special_tokens=[], seen={"spin"}, remove_collisions=False
        )
        == "SPIN"
    )
    assert (
        _determine_collision(
            token="Spin",
            is_byte=True,
            vocab=set(vocabulary),
            special_tokens=["Spin"],
            seen=set(),
            remove_collisions=False,
        )
        == "Spin"
    )

    # Unicode error
    assert (
        _determine_collision(
            token="\x80", is_byte=True, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "\x80"
    )

    # No unicode error
    assert (
        _determine_collision(
            token="\x80", is_byte=False, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "\x80"
    )

    # Decode error
    assert (
        _determine_collision(
            token="\xa1", is_byte=True, vocab=set(vocabulary), special_tokens=[], seen=set(), remove_collisions=False
        )
        == "\xa1"
    )


def test_decase() -> None:
    """Test the entire decasing procedure."""
    vocabulary = ["dog", "CAT", "Cat", "cat", "DOG", "Dog", "SPIN"]
    decased = decase_vocabulary(vocabulary, special_tokens=[], is_byte=False)
    assert decased == ["dog", "CAT", "Cat", "cat", "DOG", "Dog", "spin"]

    vocabulary = ["dog", "CAT", "Cat", "cat", "DOG", "Dog", "SPIN"]
    decased = decase_vocabulary(vocabulary, special_tokens=["SPIN"], is_byte=False)
    assert decased == ["dog", "CAT", "Cat", "cat", "DOG", "Dog", "SPIN"]

    vocabulary = ["dog", "CAT", "Cat", "cat", "DOG", "Dog", "SPIN"]
    decased = decase_vocabulary(vocabulary, special_tokens=["SPIN"], is_byte=False, remove_collisions=True)
    assert decased == ["dog", "cat", "SPIN"]



from tempfile import TemporaryDirectory
from typing import Any

import pytest
from tokenizers import Tokenizer
from tokenizers.models import BPE as TokenizersBPE
from transformers import PreTrainedTokenizerFast

from skeletoken.addedtoken import AddedTokens
from skeletoken.base import TokenizerModel
from skeletoken.common import PrependScheme
from skeletoken.merges import Merges
from skeletoken.models import BPE, ModelType, WordPiece
from skeletoken.normalizers import (
    ByteLevelNormalizer,
    LowercaseNormalizer,
    NFKCNormalizer,
    Normalizer,
    NormalizerSequence,
)
from skeletoken.padding import Padding
from skeletoken.postprocessors import (
    BertPostProcessor,
    ByteLevelPostProcessor,
    PostProcessorSequence,
    RobertaPostProcessor,
    SpecialTokenInfo,
    TemplatePostProcessor,
    Token,
    TokenSequence,
    TokenType,
)
from skeletoken.pretokenizers import (
    Behavior,
    ByteLevelPreTokenizer,
    FixedLengthPreTokenizer,
    MetaspacePreTokenizer,
    PreTokenizerSequence,
    SplitPreTokenizer,
    StringPattern,
)


def test_post_init(small_tokenizer_json: dict[str, Any]) -> None:
    """Test the post-initialization of the TokenizerModel."""
    # Remove UNK from the vocab
    small_tokenizer_json["model"]["vocab"].pop("[UNK]")
    # Remove UNK as an added token.
    small_tokenizer_json["added_tokens"] = []
    # Because the unk_token on the model is still [UNK], we will re-add it.
    model = TokenizerModel.model_validate(small_tokenizer_json)
    assert model.version == "1.0"
    assert model.model is not None
    assert isinstance(model.model, WordPiece)
    assert isinstance(model.added_tokens, AddedTokens)
    assert model.normalizer is None
    assert model.pre_tokenizer is None
    assert model.post_processor is None
    assert model.decoder is None
    assert model.unk_token == "[UNK]"
    assert "[UNK]" in model.model.vocab.vocabulary
    assert model.added_tokens.get_token("[UNK]") is not None


def test_add_addedtoken(small_tokenizer: Tokenizer) -> None:
    """Test the add_addedtoken functionality."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    model.add_addedtoken("[OSTENTATIOUS]")
    # Test if it gets added to both the model and the vocabulary.
    assert model.added_tokens.get_token("[OSTENTATIOUS]") is not None
    assert "[OSTENTATIOUS]" in model.model.vocab


def test_turn_into_addedtoken(small_tokenizer: Tokenizer) -> None:
    """Test turning a regular token into an added token."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    model.add_addedtoken("[OSTENTATIOUS]")
    # Should not crash, it is already an added token.
    model.turn_into_addedtoken("[OSTENTATIOUS]")
    # Should crash because it is not a regular token.
    with pytest.raises(ValueError):
        model.turn_into_addedtoken("bababbaa")


def test_tokenizer_model_from_tokenizer(small_tokenizer: Tokenizer) -> None:
    """Test creating a TokenizerModel from a Tokenizer instance."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)

    assert model.version == "1.0"
    assert model.model is not None
    assert isinstance(model.model, WordPiece)
    assert isinstance(model.added_tokens, AddedTokens)
    assert model.normalizer is None
    assert model.pre_tokenizer is None
    assert model.post_processor is None
    assert model.decoder is None


def test_add_pre_tokenizer(small_tokenizer: Tokenizer) -> None:
    """Test adding a pre-tokenizer to the TokenizerModel."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)

    pre_tokenizer = ByteLevelPreTokenizer(add_prefix_space=True, use_regex=True, trim_offsets=True)
    model.add_pre_tokenizer(pre_tokenizer)

    # Tests removing None and adding a pre-tokenizer
    assert model.pre_tokenizer is not None
    assert isinstance(model.pre_tokenizer, ByteLevelPreTokenizer)

    model.add_pre_tokenizer(pre_tokenizer)

    # Tests adding a second pre-tokenizer and turning it into a sequence
    assert isinstance(model.pre_tokenizer, PreTokenizerSequence)
    assert len(model.pre_tokenizer.pretokenizers) == 2

    model.add_pre_tokenizer(pre_tokenizer)

    # Tests adding a third pre-tokenizer and keeping it as a sequence
    assert isinstance(model.pre_tokenizer, PreTokenizerSequence)
    assert len(model.pre_tokenizer.pretokenizers) == 3


def test_add_normalizer(small_tokenizer: Tokenizer) -> None:
    """Test adding a normalizer to the TokenizerModel."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)

    normalizer = ByteLevelNormalizer()
    model.add_normalizer(normalizer)

    # Tests removing None and adding a normalizer
    assert model.normalizer is not None
    assert isinstance(model.normalizer, ByteLevelNormalizer)

    model.add_normalizer(normalizer)

    # Tests adding a second normalizer and turning it into a sequence
    assert isinstance(model.normalizer, NormalizerSequence)
    assert len(model.normalizer.normalizers) == 2

    model.add_normalizer(normalizer)

    # Tests adding a third normalizer and keeping it as a sequence
    assert isinstance(model.normalizer, NormalizerSequence)
    assert len(model.normalizer.normalizers) == 3


def test_add_normalizer_prefix(small_tokenizer: Tokenizer) -> None:
    """Add a normalizer prefix."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)

    normalizer: Normalizer = ByteLevelNormalizer()
    model.add_normalizer(normalizer, prefix=True)

    # Tests removing None and adding a normalizer
    assert model.normalizer is not None
    assert isinstance(model.normalizer, ByteLevelNormalizer)

    normalizer = LowercaseNormalizer()
    model.add_normalizer(normalizer, prefix=True)

    # Tests adding a second normalizer and keeping it as a sequence
    assert isinstance(model.normalizer, NormalizerSequence)
    assert len(model.normalizer.normalizers) == 2
    assert isinstance(model.normalizer.normalizers[0], LowercaseNormalizer)
    assert isinstance(model.normalizer.normalizers[1], ByteLevelNormalizer)

    normalizer = NFKCNormalizer()
    model.add_normalizer(normalizer, prefix=False)

    # Tests adding a third normalizer and keeping it as a sequence
    assert isinstance(model.normalizer, NormalizerSequence)
    assert len(model.normalizer.normalizers) == 3
    assert isinstance(model.normalizer.normalizers[0], LowercaseNormalizer)
    assert isinstance(model.normalizer.normalizers[1], ByteLevelNormalizer)
    assert isinstance(model.normalizer.normalizers[2], NFKCNormalizer)


def test_add_processor(small_tokenizer: Tokenizer) -> None:
    """Test adding a post-processor to the TokenizerModel."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)

    post_processor = ByteLevelPostProcessor(add_prefix_space=True, trim_offsets=True, use_regex=True)
    model.add_post_processor(post_processor)

    # Tests removing None and adding a post-processor
    assert model.post_processor is not None
    assert isinstance(model.post_processor, ByteLevelPostProcessor)

    model.add_post_processor(post_processor)

    # Tests adding a second post-processor and turning it into a sequence
    assert isinstance(model.post_processor, PostProcessorSequence)
    assert len(model.post_processor.processors) == 2

    model.add_post_processor(post_processor)

    # Tests adding a third post-processor and keeping it as a sequence
    assert isinstance(model.post_processor, PostProcessorSequence)
    assert len(model.post_processor.processors) == 3

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_from_pretrained(small_tokenizer: Tokenizer) -> None:
    """Test creating a TokenizerModel from a pretrained tokenizer."""
    with pytest.raises(FileNotFoundError):
        model = TokenizerModel.from_pretrained(".")

    with TemporaryDirectory() as temp_dir:
        small_tokenizer.save(f"{temp_dir}/tokenizer.json")
        model = TokenizerModel.from_pretrained(temp_dir)

        assert model.version == "1.0"
        assert model.model is not None
        assert isinstance(model.model, WordPiece)
        assert isinstance(model.added_tokens, AddedTokens)
        assert model.normalizer is None
        assert model.pre_tokenizer is None
        assert model.post_processor is None
        assert model.decoder is None

        model = TokenizerModel.from_pretrained(f"{temp_dir}/tokenizer.json")

        assert model.version == "1.0"
        assert model.model is not None
        assert isinstance(model.model, WordPiece)
        assert isinstance(model.added_tokens, AddedTokens)
        assert model.normalizer is None
        assert model.pre_tokenizer is None
        assert model.post_processor is None
        assert model.decoder is None


def test_make_greedy(small_tokenizer: Tokenizer) -> None:
    """Test whether the make greedy function works."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    model.add_pre_tokenizer(ByteLevelPreTokenizer(add_prefix_space=True, use_regex=True, trim_offsets=True))
    model.make_model_greedy()
    assert model.model.type == ModelType.WORDPIECE
    assert model.to_tokenizer()
    pre_tok_length = None
    if model.pre_tokenizer is not None:
        if isinstance(model.pre_tokenizer, PreTokenizerSequence):
            for pt in model.pre_tokenizer.pretokenizers:
                if isinstance(pt, FixedLengthPreTokenizer):
                    pre_tok_length = pt.length
        elif isinstance(model.pre_tokenizer, FixedLengthPreTokenizer):
            pre_tok_length = model.pre_tokenizer.length
    assert pre_tok_length is not None
    assert pre_tok_length == model.model.max_input_chars_per_word


def test_lowercase(small_tokenizer: Tokenizer) -> None:
    """Tests whether the model performs the lowercase test correctly."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    assert not model.lowercases_input
    model.normalizer = LowercaseNormalizer()
    assert model.lowercases_input
    assert model.lowercases_input == model.normalizer._lowercases

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_byte_normalizes(small_tokenizer: Tokenizer) -> None:
    """Tests whether the model performs byte normalization correctly."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    assert not model.transforms_into_bytes
    model.normalizer = ByteLevelNormalizer()
    assert model.transforms_into_bytes
    model.normalizer = None
    model.pre_tokenizer = ByteLevelPreTokenizer(add_prefix_space=True, use_regex=True, trim_offsets=False)
    assert model.transforms_into_bytes

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_remove_token(small_tokenizer: Tokenizer) -> None:
    """Test removing a token from the vocabulary."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    model.remove_token_from_vocabulary("a")
    with pytest.raises(ValueError):
        model.remove_token_from_vocabulary("a")
    assert "a" not in model.model.vocab.vocabulary

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_add_token(small_tokenizer: Tokenizer) -> None:
    """Test adding a token to the vocabulary."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    with pytest.raises(ValueError):
        model.add_token_to_vocabulary("a")
    model.add_token_to_vocabulary("new_token")
    assert "new_token" in model.model.vocab.vocabulary

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_replace_token(small_tokenizer: Tokenizer) -> None:
    """Test replace token interface."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    with pytest.raises(ValueError):
        model.replace_token_in_vocabulary("a", "b")
    model.replace_token_in_vocabulary("b", "x")
    assert "b" not in model.model.vocab.vocabulary
    assert "x" in model.model.vocab.vocabulary

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_replace_token_template(small_tokenizer: Tokenizer) -> None:
    """Test replace token interface."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    t = SpecialTokenInfo(id="bos", ids=[1], tokens=["a"])
    s = {"bos": t}
    tok = Token(id="bos", type_id=0, type=TokenType.SPECIAL)
    seq = Token(id="A", type_id=0, type=TokenType.SEQUENCE)
    seq2 = Token(id="B", type_id=1, type=TokenType.SEQUENCE)
    sequence = TokenSequence((tok, seq, tok))
    pair_seq = TokenSequence((tok, seq, tok, seq2, tok))
    model.post_processor = TemplatePostProcessor(special_tokens=s, single=sequence, pair=pair_seq)
    with pytest.raises(ValueError):
        model.replace_token_in_vocabulary("a", "b")
    model.replace_token_in_vocabulary("b", "x")
    assert "b" not in model.model.vocab.vocabulary
    assert "x" in model.model.vocab.vocabulary

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_decase_vocabulary(small_tokenizer: Tokenizer) -> None:
    """Test the decasing of the vocabulary."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    model.added_tokens = AddedTokens([])
    vocabulary = model.model.vocab.sorted_vocabulary
    model.decase_vocabulary()
    # This tokenizer does not assign any special tokens, so this is true.
    assert model.model.vocab.sorted_vocabulary == [x.lower() for x in vocabulary]

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_eos(small_tokenizer: Tokenizer) -> None:
    """Test getting the eos token."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    assert model.eos is None
    model.post_processor = ByteLevelPostProcessor(add_prefix_space=True, trim_offsets=True, use_regex=True)
    assert model.eos is None
    model.post_processor = None
    model.post_processor = RobertaPostProcessor(
        sep=("[SEP]", 1), cls=("[CLS]", 0), trim_offsets=True, add_prefix_space=False
    )
    assert model.eos == ["[SEP]"]
    model.post_processor = None
    model.post_processor = BertPostProcessor(sep=("[SEP]", 1), cls=("[CLS]", 0))
    assert model.eos == ["[SEP]"]
    model.post_processor = PostProcessorSequence(
        processors=[
            RobertaPostProcessor(sep=("[SEP]", 1), cls=("[CLS]", 0), trim_offsets=True, add_prefix_space=False),
            BertPostProcessor(sep=("[SEP]", 1), cls=("[CLS]", 0)),
        ]
    )
    assert model.eos is None

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_bos(small_tokenizer: Tokenizer) -> None:
    """Test getting the eos token."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    assert model.bos is None
    model.post_processor = ByteLevelPostProcessor(add_prefix_space=True, trim_offsets=True, use_regex=True)
    assert model.bos is None
    model.post_processor = None
    model.post_processor = RobertaPostProcessor(
        sep=("[SEP]", 1), cls=("[CLS]", 0), trim_offsets=True, add_prefix_space=False
    )
    assert model.bos == ["[CLS]"]
    model.post_processor = None
    model.post_processor = BertPostProcessor(sep=("[SEP]", 1), cls=("[CLS]", 0))
    assert model.bos == ["[CLS]"]
    model.post_processor = PostProcessorSequence(
        processors=[
            RobertaPostProcessor(sep=("[SEP]", 1), cls=("[CLS]", 0), trim_offsets=True, add_prefix_space=False),
            BertPostProcessor(sep=("[SEP]", 1), cls=("[CLS]", 0)),
        ]
    )
    assert model.bos is None

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_split(small_tokenizer: Tokenizer) -> None:
    """Test whether the split works correctly."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    assert not model.splits
    pretokenizer = ByteLevelPreTokenizer(add_prefix_space=True, use_regex=False, trim_offsets=True)
    model.add_pre_tokenizer(pretokenizer)
    assert not model.splits
    pretokenizer = ByteLevelPreTokenizer(add_prefix_space=True, use_regex=True, trim_offsets=True)
    model.add_pre_tokenizer(pretokenizer)
    assert model.splits

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_subword_prefix(small_tokenizer: Tokenizer) -> None:
    """Test getting the subword prefix token."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    assert model.subword_prefix == "##"

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_word_prefix(small_tokenizer: Tokenizer) -> None:
    """Test getting the word prefix token."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    assert model.word_prefix is None

    model.pre_tokenizer = ByteLevelPreTokenizer(add_prefix_space=True, use_regex=True, trim_offsets=True)
    assert model.word_prefix == "Ġ"

    model.pre_tokenizer = MetaspacePreTokenizer(replacement="▁", split=True, prepend_scheme=PrependScheme.ALWAYS)
    assert model.word_prefix == "▁"

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_replace_token_in_vocabulary(small_tokenizer: Tokenizer) -> None:
    """Test replacing a token in the vocabulary."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    model.replace_token_in_vocabulary("f", "g")
    assert model.model.vocab["g"] is not None
    assert model.added_tokens.get_token("g") is not None

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_remove_token_from_vocabulary(small_tokenizer: Tokenizer) -> None:
    """Test removing a token from the vocabulary."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    assert model.added_tokens.get_token("f") is not None
    model.remove_token_from_vocabulary("f")
    assert model.added_tokens.get_token("f") is None

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_set_unk_token(small_tokenizer: Tokenizer) -> None:
    """Test setting the unknown token."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    model.unk_token = "new_unk"
    assert model.model.vocab["new_unk"] is not None
    assert model.added_tokens.get_token("new_unk") is not None

    with pytest.raises(ValueError):
        model.unk_token = None

    assert isinstance(model.model, WordPiece)
    model.model = BPE(
        vocab=model.model.vocab,
        merges=Merges([]),
        dropout=0.0,
        unk_token="a",
        continuing_subword_prefix="",
        end_of_word_suffix="",
        fuse_unk=True,
        byte_fallback=False,
        ignore_merges=False,
    )
    model.unk_token = None

    model.unk_token = "a"
    model.unk_token = "b"
    model.unk_token = "a"

    model.unk_token = None
    model.unk_token = "OSTENTATIOUS"
    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_set_unk_token_template(small_tokenizer: Tokenizer) -> None:
    """Test setting the unknown token."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    t = SpecialTokenInfo(id="bos", ids=[1], tokens=["a"])
    s = {"bos": t}
    tok = Token(id="bos", type_id=1, type=TokenType.SPECIAL)
    seq = Token(id="A", type_id=0, type=TokenType.SEQUENCE)
    seq2 = Token(id="B", type_id=0, type=TokenType.SEQUENCE)
    sequence = TokenSequence((tok, seq, tok))
    pair_seq = TokenSequence((tok, seq, tok, seq2, tok))
    model.post_processor = TemplatePostProcessor(special_tokens=s, single=sequence, pair=pair_seq)
    model.unk_token = "a"
    model.unk_token = "b"

    assert isinstance(model.post_processor, TemplatePostProcessor)
    assert model.post_processor.special_tokens["bos"].tokens == ["b"]

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_get_padding_token(small_tokenizer: Tokenizer) -> None:
    """Get the padding token from the tokenizer model."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    assert model.padding is None
    assert model.pad_token is None

    model.padding = Padding(pad_token="[PAD]", pad_id=3, pad_type_id=0)
    assert model.pad_token == "[PAD]"

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_set_padding_token(small_tokenizer: Tokenizer) -> None:
    """Set the padding token for the tokenizer model."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    model.pad_token = "[PAD]"
    assert model.added_tokens.get_token("[PAD]") is not None
    model.pad_token = None
    model.pad_token = "OSTENTATIOUS"
    assert model.added_tokens.get_token("OSTENTATIOUS") is not None
    assert model.added_tokens.get_token("[PAD]") is not None
    model.added_tokens = AddedTokens([])
    model.pad_token = "OSTENTATIOUS"
    assert model.added_tokens.get_token("OSTENTATIOUS") is not None
    model.pad_token = "OSTENTATIOUS"
    assert model.added_tokens.get_token("OSTENTATIOUS") is not None
    model.pad_token = "FUN"
    assert model.added_tokens.get_token("FUN") is not None

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_from_transformers(transformers_tokenizer: PreTrainedTokenizerFast) -> None:
    """Test creating a TokenizerModel from a transformers tokenizer."""
    model = TokenizerModel.from_transformers_tokenizer(transformers_tokenizer)
    assert model.version == "1.0"
    assert model.model is not None
    assert isinstance(model.model, WordPiece)
    assert isinstance(model.added_tokens, AddedTokens)

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_from_transformers_missing_unk(transformers_tokenizer: PreTrainedTokenizerFast) -> None:
    """Test creating a TokenizerModel from a transformers tokenizer."""
    transformers_tokenizer.SPECIAL_TOKENS_ATTRIBUTES.remove("unk_token")
    model = TokenizerModel.from_transformers_tokenizer(transformers_tokenizer)
    assert model.version == "1.0"
    assert model.model is not None
    assert isinstance(model.added_tokens, AddedTokens)

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_from_transformers_missing_pad(transformers_tokenizer: PreTrainedTokenizerFast) -> None:
    """Test creating a TokenizerModel from a transformers tokenizer."""
    transformers_tokenizer.SPECIAL_TOKENS_ATTRIBUTES.remove("pad_token")
    model = TokenizerModel.from_transformers_tokenizer(transformers_tokenizer)
    assert model.version == "1.0"
    assert model.model is not None
    assert isinstance(model.model, WordPiece)
    assert isinstance(model.added_tokens, AddedTokens)

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_missing_unk_model(transformers_tokenizer: PreTrainedTokenizerFast) -> None:
    """Test creating a TokenizerModel from a transformers tokenizer."""
    transformers_tokenizer._tokenizer.model = TokenizersBPE(
        vocab={"a": 0, "b": 1, "c": 2, "d": 3, "e": 4, "f": 5, " ": 6}, merges=[], unk_token=None
    )  # type: ignore
    transformers_tokenizer._tokenizer.enable_padding(pad_id=3, pad_token="[PAD]")

    model = TokenizerModel.from_transformers_tokenizer(transformers_tokenizer)
    assert model.version == "1.0"
    assert model.model is not None
    assert isinstance(model.added_tokens, AddedTokens)

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def test_model_delta(small_tokenizer: Tokenizer) -> None:
    """Test the model delta functionality."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    model.add_token_to_vocabulary("new_token")
    model.remove_token_from_vocabulary("a")
    model.replace_token_in_vocabulary("b", "x")
    model.add_addedtoken("[ADDED]")
    model.unk_token = "new_unk"
    model.pad_token = "[PAD]"
    delta = model.model_delta
    assert delta.token_mapping == {7: 6, 8: 7, 9: 8}
    assert delta.new_vocabulary_size == 13
    assert delta.new_tokens == {"x": 5, "[ADDED]": 11, "new_token": 10, "new_unk": 12, "f": 9}


def test_model_same_unk_and_pad(small_tokenizer: Tokenizer) -> None:
    """Test the model delta functionality."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    model.add_pre_tokenizer(
        SplitPreTokenizer(pattern=StringPattern(String=" "), behavior=Behavior.ISOLATED, invert=False)
    )
    model = model.make_model_greedy()
    model.unk_token = "[UNK]"
    unk = model.added_tokens.get_token("[UNK]")
    length_tokens = len(model.added_tokens)
    assert unk is not None
    unk_id = unk.id
    model.pad_token = "[UNK]"
    pad = model.added_tokens.get_token("[UNK]")
    assert pad is not None
    pad_id = pad.id
    assert unk_id == pad_id
    assert len(model.added_tokens) == length_tokens

    assert model.unk_token == "[UNK]"
    assert model.pad_token == "[UNK]"

    tokenizer = model.to_tokenizer()
    encoded = tokenizer.encode("abcdef")
    assert encoded.tokens == ["[UNK]", "f"]
    assert encoded.ids == [2, 10]


def test_token_to_id_and_id_to_token(small_tokenizer: Tokenizer) -> None:
    """Test the tokens_to_ids and ids_to_tokens methods."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    tokens = ["a", "b", "c", "d", "f"]
    ids = model.tokens_to_ids(tokens)
    assert ids == [5, 6, 7, 8, 10]
    recovered_tokens = model.ids_to_tokens(ids)
    assert recovered_tokens == tokens

    tokenizer = model.to_tokenizer()
    encoded = tokenizer.encode_batch(["a", "b", "c", "d", "f"])
    assert [enc.tokens[0] for enc in encoded] == tokens
    assert [enc.ids[0] for enc in encoded] == ids


def test_to_transformers(small_tokenizer: Tokenizer) -> None:
    """Test saving and loading a tokenizer model."""
    model = TokenizerModel.from_tokenizer(small_tokenizer)
    transformers_tokenizer = model.to_transformers()
    assert transformers_tokenizer.pad_token is None
    assert transformers_tokenizer.unk_token == "[UNK]"

    model.pad_token = "[PAD]"
    transformers_tokenizer = model.to_transformers()
    assert transformers_tokenizer.pad_token == "[PAD]"

    model.post_processor = RobertaPostProcessor(
        sep=("[SEP]", 1), cls=("[CLS]", 0), trim_offsets=True, add_prefix_space=False
    )
    assert model.eos == ["[SEP]"]
    assert model.bos == ["[CLS]"]
    transformers_tokenizer = model.to_transformers()
    assert transformers_tokenizer.eos_token == "[SEP]"
    assert transformers_tokenizer.bos_token == "[CLS]"

    model.post_processor = TemplatePostProcessor(
        special_tokens={
            "sep": SpecialTokenInfo(id="sep", ids=[1, 2], tokens=["[SEP]", "</s>"]),
            "cls": SpecialTokenInfo(id="cls", ids=[0, 1], tokens=["[CLS]", "<s>"]),
        },
        single=TokenSequence(
            (
                Token(id="cls", type_id=0, type=TokenType.SPECIAL),
                Token(id="A", type_id=0, type=TokenType.SEQUENCE),
                Token(id="sep", type_id=1, type=TokenType.SPECIAL),
            )
        ),
        pair=TokenSequence(
            (
                Token(id="cls", type_id=0, type=TokenType.SPECIAL),
                Token(id="A", type_id=0, type=TokenType.SEQUENCE),
                Token(id="sep", type_id=1, type=TokenType.SPECIAL),
                Token(id="B", type_id=1, type=TokenType.SEQUENCE),
                Token(id="sep", type_id=1, type=TokenType.SPECIAL),
            )
        ),
    )
    assert model.eos == ["[SEP]", "</s>"]
    assert model.bos == ["[CLS]", "<s>"]

    transformers_tokenizer = model.to_transformers()
    assert transformers_tokenizer.eos_token is None
    assert transformers_tokenizer.bos_token is None



import re
from enum import Enum

import regex
from pydantic import BaseModel


class StringPattern(BaseModel):
    """A string pattern for use in a replace."""

    String: str


class RegexPattern(BaseModel):
    """A regex pattern for use in a replace."""

    Regex: str


def coerce_string_regex_pattern(
    v: str | regex.Pattern | re.Pattern | dict | StringPattern | RegexPattern,
) -> dict | StringPattern | RegexPattern:
    """Helper function that turns a string or regex pattern into the appropriate dict."""
    # Users can pass: str, compiled regex, or the tagged dict forms
    if isinstance(v, (StringPattern, RegexPattern)):
        return v
    if isinstance(v, str):
        return {"String": v}
    if isinstance(v, (regex.Pattern, re.Pattern)):
        return {"Regex": v.pattern}
    if isinstance(v, dict) and (("String" in v) or ("Regex" in v)):
        return v
    raise TypeError("pattern must be a string, a compiled regex, or a dict like {'String': ...} / {'Regex': ...}")


class PrependScheme(str, Enum):
    """
    The prepend scheme used in metaspace tokenizers.

    This enum governs the behavior of the metaspace prepend.

    ALWAYS: prepend all sequences with a metaspace.
    NEVER: never prepend. The behavior of this depends on other tokenizers.
        If Metaspace is the only tokenizer, this only governs the behavior of
        the first token in the sequence, otherwise it also affects other tokens.
    FIRST: Only prepend the first token in a sequence with a metaspace.
    """

    ALWAYS = "always"
    NEVER = "never"
    FIRST = "first"


class Behavior(str, Enum):
    """
    Behavior for split.

    For pattern "a", the behavior is as follows:

    REMOVED: removes the character from the sequence. "bab" -> "b", "b"
    ISOLATED: splits the character as a separate token. "bab" -> "b", "a", "b"
    MERGED_WITH_PREVIOUS: merges with the token before it. "bab" -> "ba", "b"
    MERGED_WITH_NEXT: merges with the following token. "bab" -> "b", "ab"
    CONTIGUOUS: merges contiguous blocks of patterns: "baaab" -> "b", "aaa", "b"
    """

    REMOVED = "Removed"
    ISOLATED = "Isolated"
    MERGED_WITH_PREVIOUS = "MergedWithPrevious"
    MERGED_WITH_NEXT = "MergedWithNext"
    CONTIGUOUS = "Contiguous"



from __future__ import annotations

from enum import Enum
from typing import Annotated, Any, Literal

from pydantic import BaseModel, Field, field_validator

from skeletoken.common import PrependScheme, RegexPattern, StringPattern, coerce_string_regex_pattern


class DecoderType(str, Enum):
    BPEDECODER = "BPEDecoder"
    BYTEFALLBACK = "ByteFallback"
    BYTELEVEL = "ByteLevel"
    CTC = "CTC"
    FUSE = "Fuse"
    METASPACE = "Metaspace"
    REPLACE = "Replace"
    STRIP = "Strip"
    WORDPIECE = "WordPiece"
    SEQUENCE = "Sequence"


class DecoderSequence(BaseModel):
    """A sequence of decoders."""

    type: Literal[DecoderType.SEQUENCE] = DecoderType.SEQUENCE
    decoders: list[Decoder]


class BPEDecoder(BaseModel):
    """
    A legacy BPE decoder.

    The `suffix` is used to determine the end of a BPE token.
    For example, if the suffix is "a", then "baba" will be decoded as "bb".

    Attributes
    ----------
    type : Literal[DecoderType.BPEDECODER]
        The type of the decoder. This is always "BPEDecoder".
    suffix : str
        The suffix to use for BPE tokenization.

    """

    type: Literal[DecoderType.BPEDECODER] = DecoderType.BPEDECODER
    suffix: str


class ByteFallbackDecoder(BaseModel):
    """
    A ByteFallback decoder is used to handle byte-level tokens.

    It acts as a fallback for weird bytes, and replaces them with �, much like
    'utf-8' decoding does.

    """

    type: Literal[DecoderType.BYTEFALLBACK] = DecoderType.BYTEFALLBACK


class ByteLevelDecoder(BaseModel):
    """
    A ByteLevel decoder is used for byte-level tokenization.

    This decoder implements the inverse of the ByteLevel pretokenizer.

    Attributes
    ----------
    type : Literal[DecoderType.BYTELEVEL]
        The type of the decoder. This is always "ByteLevel".
    add_prefix_space : bool
        Whether to add a space before the first token. This leads to more consistent
        behavior for sentence-initial tokens, and is recommended to be set to True.
    trim_offsets : bool
        Whether to trim the offsets of the tokens.
    use_regex : bool
        Whether the regex was used for splitting. If this is not set, no splitting was performed. The
        regex used is the same as in the ByteLevel pretokenizer.

    """

    type: Literal[DecoderType.BYTELEVEL] = DecoderType.BYTELEVEL
    add_prefix_space: bool
    trim_offsets: bool
    use_regex: bool


class CTCDecoder(BaseModel):
    """
    A CTC decoder is used for connectionist temporal classification.

    It removes any contiguous duplicates, e.g., "hh_ee_ll_ll_oo" -> "hello",
    and removes the padding token, e.g., "h h _ e e l l l _ o o" -> "hello".
    The word delimiter is replaced by a space between words, e.g., "hello|world" -> "hello | world".

    Attributes
    ----------
    type : Literal[DecoderType.CTC]
        The type of the decoder. This is always "CTC".
    pad_token : str
        The padding token to remove.
    word_delimiter_token : str
        The token used to separate words.
    cleanup : bool
        If set, it will clean up the output by removing some artifacts.

    """

    type: Literal[DecoderType.CTC] = DecoderType.CTC
    pad_token: str
    word_delimiter_token: str
    cleanup: bool


class FuseDecoder(BaseModel):
    """
    A Fuse decoder just merges tokens.

    e.g., "un" + "known" -> "unknown"

    """

    type: Literal[DecoderType.FUSE] = DecoderType.FUSE


class MetaspaceDecoder(BaseModel):
    """
    A Metaspace decoder is used for metaspace tokenization.

    This decoder inverts the metaspace tokenization process.

    Attributes
    ----------
    type : Literal[DecoderType.METASPACE]
        The type of the decoder. This is always "Metaspace".
    replacement : str
        The string to replace the metaspace tokens with.
    prepend_scheme : PrependScheme
        The scheme to use for prepending the replacement string.
    split : bool
        Should be set if the original Metaspace tokenizer performed splitting.

    """

    type: Literal[DecoderType.METASPACE] = DecoderType.METASPACE
    replacement: str
    prepend_scheme: PrependScheme
    split: bool


class ReplaceDecoder(BaseModel):
    """
    A Replace decoder replaces a pattern in the input text with a given content.

    It can be seen as the inverse of the Replace pretokenizer, but doesn't need to be the exact inverse.

    """

    type: Literal[DecoderType.REPLACE] = DecoderType.REPLACE
    pattern: StringPattern | RegexPattern
    content: str

    @field_validator("pattern", mode="before")
    @classmethod
    def _coerce_content(cls, v: Any) -> StringPattern | RegexPattern | dict:
        return coerce_string_regex_pattern(v)


class StripDecoder(BaseModel):
    """
    A Strip decoder strips characters from the input text.

    This decoder removes a specific content from the start and/or end of the input text.
    The start and stop indices are used to determine where to strip the content.
    If the start index is higher than the stop index, it will strip from the end of the content.

    e.g., start 0 and end 1 will strip the first character,
    while start 1 and end 0 will strip the last character.

    Attributes
    ----------
    type : Literal[DecoderType.STRIP]
        The type of the decoder. This is always "Strip".
    content : str
        The content to strip.
    start : int
        The start index to strip from.
    stop : int
        The stop index to strip to.

    """

    type: Literal[DecoderType.STRIP] = DecoderType.STRIP
    content: str
    start: int
    stop: int


class WordPieceDecoder(BaseModel):
    """
    A WordPiece decoder is used for WordPiece tokenization.

    This decoder implements the inverse of the WordPiece pretokenizer.

    Attributes
    ----------
    type : Literal[DecoderType.WORDPIECE]
        The type of the decoder. This is always "WordPiece".
    prefix : str
        The subword prefix to use for WordPiece tokenization. This is usually
        '##'.
    cleanup : bool
        If True, it will clean up the output by removing some artifacts.

    """

    type: Literal[DecoderType.WORDPIECE] = DecoderType.WORDPIECE
    prefix: str
    cleanup: bool


Decoder = (
    BPEDecoder
    | ByteFallbackDecoder
    | ByteLevelDecoder
    | CTCDecoder
    | FuseDecoder
    | MetaspaceDecoder
    | ReplaceDecoder
    | StripDecoder
    | WordPieceDecoder
    | DecoderSequence
)
DecoderDiscriminator = Annotated[Decoder, Field(discriminator="type")]



from __future__ import annotations

import logging
from enum import Enum
from typing import Annotated, Any, Literal

from pydantic import BaseModel, Field, SerializationInfo, model_serializer, model_validator

logger = logging.getLogger(__name__)


class PostProcessorType(str, Enum):
    SEQUENCE = "Sequence"
    BERT_PROCESSING = "BertProcessing"
    BYTE_LEVEL = "ByteLevel"
    ROBERTA_PROCESSING = "RobertaProcessing"
    TEMPLATE_PROCESSING = "TemplateProcessing"


class PostProcessorSequence(BaseModel):
    """A sequence of postprocessors."""

    type: Literal[PostProcessorType.SEQUENCE] = PostProcessorType.SEQUENCE
    processors: list[PostProcessor]


class BertPostProcessor(BaseModel):
    """
    The BERT postprocessor.

    This adds the SEP and CLS tokens to the sequence.
    Note that this processor is actually never used, even BERT uses
    the TemplatePostProcessor, see below.

    Attributes
    ----------
    type : Literal[PostProcessorType.BERT_PROCESSING]
        The type of the postprocessor. This is always "BertProcessing".
    sep : tuple[str, int]
        The SEP token and its token id.
    cls : tuple[str, int]
        The CLS token and its token id.

    """

    type: Literal[PostProcessorType.BERT_PROCESSING] = PostProcessorType.BERT_PROCESSING
    sep: tuple[str, int]
    cls: tuple[str, int]


class ByteLevelPostProcessor(BaseModel):
    """The ByteLevelPostProcessor. This adds the prefix space, and trims the offsets."""

    type: Literal[PostProcessorType.BYTE_LEVEL] = PostProcessorType.BYTE_LEVEL
    add_prefix_space: bool
    trim_offsets: bool
    use_regex: bool


class RobertaPostProcessor(BaseModel):
    """
    The RoBERTa postprocessor.

    This adds the SEP and CLS tokens to the sequence.
    Note that this processor is actually never used, even RoBERTa uses
    the TemplatePostProcessor, see below.

    Attributes
    ----------
    type : Literal[PostProcessorType.ROBERTA_PROCESSING]
        The type of the postprocessor. This is always "RobertaProcessing".
    sep : tuple[str, int]
        The SEP token and its token id.
    cls : tuple[str, int]
        The CLS token and its token id.
    trim_offsets : bool
        Whether to trim the offsets of the tokens.
    add_prefix_space : bool
        Whether to add a space before the first token. This leads to more consistent
        behavior for sentence-initial tokens, and is recommended to be set to True.

    """

    type: Literal[PostProcessorType.ROBERTA_PROCESSING] = PostProcessorType.ROBERTA_PROCESSING
    sep: tuple[str, int]
    cls: tuple[str, int]
    trim_offsets: bool
    add_prefix_space: bool


class TokenType(str, Enum):
    SPECIAL = "SpecialToken"
    SEQUENCE = "Sequence"


class Token(BaseModel):
    id: str
    type_id: int
    type: TokenType

    @model_validator(mode="before")
    @classmethod
    def parse_format(cls, v) -> dict:
        """Parse either {'SpecialToken': {...}} or {'Sequence': {...}}."""
        if isinstance(v, dict):
            # Two types:
            # 1. {"Sequence": {... }} or {"SpecialToken": {...}}
            # 2. {"id": ..., "type_id": ..., "type": ...}
            if len(v) == 1:
                t = next(iter(v))
                v = v[t]
            else:
                t = v["type"]
            inner = {**v}
            inner["type"] = TokenType(t)
            return inner
        raise TypeError("Token must be a dict or a Token instance.")

    @model_serializer
    def serializer(self, info: SerializationInfo) -> dict[str, Any]:
        """Serialization to either {'SpecialToken': {...}} or {'Sequence': {...}}."""
        data = {"id": self.id, "type_id": self.type_id}
        return {self.type: data}


class SpecialTokenInfo(BaseModel):
    """Information about a special token used in a template post-processor."""

    id: str
    ids: list[int]
    tokens: list[str]

    def model_post_init(self, __context: dict) -> None:
        """Validates that ids and tokens have the same length."""
        if len(self.ids) != len(self.tokens):
            logger.warning("ids and tokens must have the same length. Ids: %s, tokens: %s", self.ids, self.tokens)


# Simple type alias for a sequence of tokens for a template post-processor.
TokenSequence = tuple[Token, ...]


class TemplatePostProcessor(BaseModel):
    type: Literal[PostProcessorType.TEMPLATE_PROCESSING] = PostProcessorType.TEMPLATE_PROCESSING
    single: TokenSequence
    pair: TokenSequence
    special_tokens: dict[str, SpecialTokenInfo]

    def model_post_init(self, __context: dict) -> None:
        """Validates that all special tokens in single and pair are defined in special_tokens."""
        for token_name, t in self.special_tokens.items():
            if token_name != t.id:
                raise ValueError(f"Special token name {token_name} does not match its id {t.id}.")
        num_types = _count_and_check_types(self.single, self.special_tokens)
        if num_types != 1:
            raise ValueError("Single sequence template can only have one sequence type.")
        num_types = _count_and_check_types(self.pair, self.special_tokens)
        if num_types != 2:
            raise ValueError("Pair sequence template can only have two sequence types.")


def _count_and_check_types(tokens: TokenSequence, special_tokens: dict[str, SpecialTokenInfo]) -> int:
    """Count the number of unique type_ids in a sequence of tokens."""
    sequence_tokens = 0
    unique_types = set()
    for token in tokens:
        if token.type == TokenType.SEQUENCE:
            sequence_tokens += 1
            unique_types.add(token.id)
        if token.type == TokenType.SPECIAL and token.id not in special_tokens:
            raise ValueError(f"Special token {token.id} is not defined in special_tokens.")
    if sequence_tokens != len(unique_types):
        raise ValueError("All sequence tokens must have a unique id.")
    return sequence_tokens


PostProcessor = (
    BertPostProcessor | ByteLevelPostProcessor | RobertaPostProcessor | TemplatePostProcessor | PostProcessorSequence
)
PostProcessorDiscriminator = Annotated[PostProcessor, Field(discriminator="type")]


def get_bos_token_from_post_processor(post_processor: PostProcessor) -> list[str] | None:
    """Get the beginning-of-sequence token from a post-processor."""
    if isinstance(post_processor, PostProcessorSequence):
        return None
    if isinstance(post_processor, ByteLevelPostProcessor):
        return None
    if isinstance(post_processor, (RobertaPostProcessor, BertPostProcessor)):
        return [post_processor.cls[0]]
    if isinstance(post_processor, TemplatePostProcessor):
        single_encoding = post_processor.single
        if single_encoding[0].type != TokenType.SPECIAL:
            return None
        identifier = single_encoding[0].id
        return post_processor.special_tokens[identifier].tokens


def get_eos_token_from_post_processor(post_processor: PostProcessor) -> list[str] | None:
    """Get the end-of-sequence token from a post-processor."""
    if isinstance(post_processor, PostProcessorSequence):
        return None
    if isinstance(post_processor, ByteLevelPostProcessor):
        return None
    if isinstance(post_processor, (RobertaPostProcessor, BertPostProcessor)):
        return [post_processor.sep[0]]
    if isinstance(post_processor, TemplatePostProcessor):
        single_encoding = post_processor.single
        if single_encoding[-1].type != TokenType.SPECIAL:
            return None
        identifier = single_encoding[-1].id
        return post_processor.special_tokens[identifier].tokens


def maybe_replace_token_in_post_processor(
    old_token: str, new_token: str, index: int, post_processor: PostProcessorDiscriminator
) -> PostProcessor:
    """
    Replace a token in a post-processor, if it exists.

    Parameters
    ----------
    old_token : str
        The token to replace.
    new_token : str
        The new token to insert.
    index : int
        The new index to insert.
    post_processor : PostProcessorDiscriminator
        The post-processor to replace the token in.

    Returns
    -------
    PostProcessor
        The post-processor with the token replaced, if it existed.

    """
    post_processor = post_processor.model_copy(deep=True)
    if isinstance(post_processor, PostProcessorSequence):
        return PostProcessorSequence(
            processors=[
                maybe_replace_token_in_post_processor(old_token, new_token, index, p) for p in post_processor.processors
            ]
        )
    if isinstance(post_processor, ByteLevelPostProcessor):
        # Has no tokens.
        return post_processor
    if isinstance(post_processor, (RobertaPostProcessor, BertPostProcessor)):
        # cls and sep can be the same token.
        if old_token == post_processor.cls[0]:
            post_processor.cls = (new_token, index)
        if old_token == post_processor.sep[0]:
            post_processor.sep = (new_token, index)
    if isinstance(post_processor, TemplatePostProcessor):
        for special_token in post_processor.special_tokens.values():
            new_tokens = []
            new_ids = []
            for token, id in zip(special_token.tokens, special_token.ids, strict=True):
                if token == old_token:
                    new_tokens.append(new_token)
                    new_ids.append(index)
                else:
                    new_tokens.append(token)
                    new_ids.append(id)
            special_token.tokens = new_tokens
            special_token.ids = new_ids

    return post_processor



from __future__ import annotations

from enum import Enum
from typing import Annotated, Any, Literal

from pydantic import BaseModel, Field, PrivateAttr, field_validator

from skeletoken.common import Behavior, PrependScheme, RegexPattern, StringPattern, coerce_string_regex_pattern


class PreTokenizerType(str, Enum):
    BERT_PRETOKENIZER = "BertPreTokenizer"
    BYTELEVEL = "ByteLevel"
    CHARDELIMITERSPLIT = "CharDelimiterSplit"
    DIGITS = "Digits"
    FIXEDLENGTH = "FixedLength"
    METASPACE = "Metaspace"
    PUNCTUATION = "Punctuation"
    SPLIT = "Split"
    WHITESPACE = "Whitespace"
    WHITESPACESPLIT = "WhitespaceSplit"
    UNICODESCRIPTS = "UnicodeScripts"
    SEQUENCE = "Sequence"


class PreTokenizerSequence(BaseModel):
    """A sequence of pretokenizers to be applied in order."""

    type: Literal[PreTokenizerType.SEQUENCE] = PreTokenizerType.SEQUENCE
    pretokenizers: list[PreTokenizer]

    @property
    def _byte_pretokenizes(self) -> bool:
        return any(x._byte_pretokenizes for x in self.pretokenizers)

    @property
    def _splits(self) -> bool:
        return any(x._splits for x in self.pretokenizers)


class BasePretokenizer(BaseModel):
    """A base class for pretokenizers."""

    _byte_pretokenizes: bool = PrivateAttr(default=False)

    @property
    def _splits(self) -> bool:
        """Most pretokenizers split."""
        return True


class BertPreTokenizer(BasePretokenizer):
    """
    The BERT pretokenizer.

    This pretokenizer splits on spaces and punctuation.
    Each occurrence of a punctuation character counts separately,

    so: ",,," will be pretokenized as [",", ",", ","], despite these
    not being separated by spaces.

    Attributes
    ----------
    type : Literal[PreTokenizerType.BERT_PRETOKENIZER]
        The type of the pretokenizer. This is always "BertPreTokenizer".

    """

    type: Literal[PreTokenizerType.BERT_PRETOKENIZER] = PreTokenizerType.BERT_PRETOKENIZER


class ByteLevelPreTokenizer(BasePretokenizer):
    """
    The pretokenizer used for BPE.

    This pretokenizer converts your input text into a sequence of byte-level tokens.
    These byte level tokens are not exactly the standard byte ranges, but a remapped
    version.

    This pretokenizer only splits, i.e., creates pre-tokens, if use_regex is set. Otherwise,
    it creates a single token, and should be accompanied by another pretokenizer.

    Attributes
    ----------
    type : Literal[PreTokenizerType.BYTELEVEL]
        The type of the pretokenizer. This is always "ByteLevel".
    add_prefix_space : bool
        Whether to add a space before the first token. This is desirable because it
        creates more correspondence between tokens.
    use_regex: bool
        Whether to use regex for splitting. If this is not set, no splitting is performed.
    trim_offsets: bool
        Doesn't do anything
    _byte_pretokenizes: bool
        Whether this pretokenizer produces byte-level tokens. This is always True.

    """

    type: Literal[PreTokenizerType.BYTELEVEL] = PreTokenizerType.BYTELEVEL
    add_prefix_space: bool
    use_regex: bool
    trim_offsets: bool
    _byte_pretokenizes: bool = PrivateAttr(default=True)

    @property
    def _splits(self) -> bool:
        """Only splits if use_regex is True."""
        return self.use_regex


class CharDelimiterSplitPreTokenizer(BasePretokenizer):
    """
    Pretokenizes on a specific delimiter.

    The delimiter should be a single character, and can't be a regular expression.

    Attributes
    ----------
    type : Literal[PreTokenizerType.CHARDELIMITERSPLIT]
        The type of the pretokenizer. This is always "CharDelimiterSplit".
    delimiter : str
        The character to split on.

    """

    type: Literal[PreTokenizerType.CHARDELIMITERSPLIT] = PreTokenizerType.CHARDELIMITERSPLIT
    delimiter: str


class DigitsPreTokenizer(BasePretokenizer):
    """
    Split on digits.

    This tokenizer splits digits, and possibly sequences of digits, into tokens.

    Attributes
    ----------
    type : Literal[PreTokenizerType.DIGITS]
        The type of the pretokenizer. This is always "Digits".
    individual_digits : bool
        If set, this splits on individual digits. If not, this splits on groups of digits.
        e.g. "111" -> ["1", "1", "1"] if this is set.

    """

    #: The type of the pretokenizer.
    type: Literal[PreTokenizerType.DIGITS] = PreTokenizerType.DIGITS
    individual_digits: bool


class FixedLengthPreTokenizer(BasePretokenizer):
    """
    Pretokenizes into fixed length sequences.

    Attributes
    ----------
    type : Literal[PreTokenizerType.FIXEDLENGTH]
        The type of the pretokenizer. This is always "FixedLength".
    length : int
        The length of each sequence.

    """

    type: Literal[PreTokenizerType.FIXEDLENGTH] = PreTokenizerType.FIXEDLENGTH
    length: int


class MetaspacePreTokenizer(BasePretokenizer):
    """
    The metaspace pretokenizer.

    This tokenizer replaces the space character by a single character. This is usually
    `▁` (U+2581).

    Attributes
    ----------
    type : Literal[PreTokenizerType.METASPACE]
        The type of the pretokenizer. This is always "Metaspace".
    replacement : str
        The character to replace spaces with.
    prepend_scheme : PrependScheme
        The scheme to use for prepending the sequence.
        If this is set to `FIRST`, it will prepend the replacement character to the
        first token in the sequence.
    split : bool
        Whether to split the input text into tokens.
        The split is done on the specified replacement character, but the split character
        is kept during splitting, and prepended to the token.

    """

    type: Literal[PreTokenizerType.METASPACE] = PreTokenizerType.METASPACE
    replacement: str
    prepend_scheme: PrependScheme
    split: bool

    @property
    def _splits(self) -> bool:
        """Splits if split is set to True."""
        return self.split


class PunctuationPreTokenizer(BasePretokenizer):
    """
    Splits on punctuation, e.g., most punctuation characters.

    Note that the behavior for this is weird for unicode punctuation.

    Attributes
    ----------
    type : Literal[PreTokenizerType.PUNCTUATION]
        The type of the pretokenizer. This is always "Punctuation".
    behavior : Behavior
        See the Behavior docstring for how this functions.

    """

    type: Literal[PreTokenizerType.PUNCTUATION] = PreTokenizerType.PUNCTUATION
    behavior: Behavior


class SplitPreTokenizer(BasePretokenizer):
    """
    Split a sequence on a specified pattern.

    The specified pattern can either be a regex or a string. The string and regex
    can be specified as follows:

    {"String": my_string}
    or
    {"Regex": my_regex}

    Or can be added directly as a string or a re.regex or regex.regex Pattern object.

    Attributes
    ----------
    type : Literal[PreTokenizerType.SPLIT]
        The type of the pretokenizer. This is always "Split".
    pattern : StringPattern | RegexPattern
        The pattern to split on.
        If you use a regex, it should be a valid regex pattern.
    behavior : Behavior
        The behavior to use when splitting. See the docstring of Behavior
        for the different possibilities.
    invert : bool
        Whether to invert the pattern. The invert is ignored when a non-Regex, i.e.,
        a string pattern, is used.

    """

    type: Literal[PreTokenizerType.SPLIT] = PreTokenizerType.SPLIT
    pattern: StringPattern | RegexPattern
    behavior: Behavior
    invert: bool

    @field_validator("pattern", mode="before")
    @classmethod
    def _coerce_pattern(cls, v: Any) -> StringPattern | RegexPattern | dict:
        return coerce_string_regex_pattern(v)


class WhitespacePreTokenizer(BasePretokenizer):
    r"""
    Splits using the following regex: "\w+|[^\w\s]+".

    This can be seen as a standard word-based splitter, very similar to the
    one used in older tokenizers.

    """

    type: Literal[PreTokenizerType.WHITESPACE] = PreTokenizerType.WHITESPACE


class WhitespaceSplitPreTokenizer(BasePretokenizer):
    """
    Literally only splits on whitespace.

    This isn't really desirable, and can be considered a variant of CharDelimiterSplit.

    Attributes
    ----------
    type : Literal[PreTokenizerType.WHITESPACESPLIT]
        The type of the pretokenizer. This is always "WhitespaceSplit".

    """

    type: Literal[PreTokenizerType.WHITESPACESPLIT] = PreTokenizerType.WHITESPACESPLIT


class UnicodeScriptsPreTokenizer(BasePretokenizer):
    """
    Splits when encountering a new Unicode script.

    This is interesting when training multilingual tokenizers, because this guarantees
    that the tokenizer will split on script and language boundaries.

    Attributes
    ----------
    type : Literal[PreTokenizerType.UNICODESCRIPTS]
        The type of the pretokenizer. This is always "UnicodeScripts".

    Examples
    --------
    >>> s = "hello, ごきげんよう?"
    >>> tokenizer = UnicodeScriptsPreTokenizer()
    >>> tokenizer.pre_tokenize_str(s)
    [('hello', (0, 5)), (', ', (5, 7)), ('ごきげんよう', (7, 13)), ('?', (13, 14))]

    """

    type: Literal[PreTokenizerType.UNICODESCRIPTS] = PreTokenizerType.UNICODESCRIPTS


PreTokenizer = (
    BertPreTokenizer
    | ByteLevelPreTokenizer
    | CharDelimiterSplitPreTokenizer
    | DigitsPreTokenizer
    | FixedLengthPreTokenizer
    | MetaspacePreTokenizer
    | PunctuationPreTokenizer
    | SplitPreTokenizer
    | WhitespacePreTokenizer
    | WhitespaceSplitPreTokenizer
    | UnicodeScriptsPreTokenizer
    | PreTokenizerSequence
)
PreTokenizerDiscriminator = Annotated[PreTokenizer, Field(discriminator="type")]


def get_metaspace(pre_tokenizer: PreTokenizerDiscriminator) -> str | None:
    """Get the metaspace token from a pre-tokenizer."""
    if isinstance(pre_tokenizer, PreTokenizerSequence):
        for pt in pre_tokenizer.pretokenizers:
            if result := get_metaspace(pt):
                return result
    elif isinstance(pre_tokenizer, MetaspacePreTokenizer):
        return pre_tokenizer.replacement
    return None



from typing import Any

import pytest

from skeletoken.base import TokenizerModel
from skeletoken.common import PrependScheme, StringPattern
from skeletoken.normalizers import (
    BertNormalizer,
    ByteLevelNormalizer,
    LowercaseNormalizer,
    NFCNormalizer,
    NFDNormalizer,
    NFKCNormalizer,
    NFKDNormalizer,
    NmtNormalizer,
    Normalizer,
    NormalizerSequence,
    NormalizerType,
    PrecompiledNormalizer,
    PrependNormalizer,
    ReplaceNormalizer,
    StripAccentsNormalizer,
    StripNormalizer,
)


def _get_default_normalizer(normalizer_type: NormalizerType) -> Normalizer:  # noqa: C901
    """Helper function to get the default instantiation of a normalizer."""
    if normalizer_type == NormalizerType.BYTELEVEL:
        return ByteLevelNormalizer()
    elif normalizer_type == NormalizerType.BERTNORMALIZER:
        return BertNormalizer(clean_text=True, handle_chinese_chars=True, strip_accents=None, lowercase=True)
    elif normalizer_type == NormalizerType.LOWERCASE:
        return LowercaseNormalizer()
    elif normalizer_type == NormalizerType.NFC:
        return NFCNormalizer()
    elif normalizer_type == NormalizerType.NFD:
        return NFDNormalizer()
    elif normalizer_type == NormalizerType.NFKC:
        return NFKCNormalizer()
    elif normalizer_type == NormalizerType.NFKD:
        return NFKDNormalizer()
    elif normalizer_type == NormalizerType.NMT:
        return NmtNormalizer()
    elif normalizer_type == NormalizerType.PREPEND:
        return PrependNormalizer(prepend=PrependScheme.FIRST)
    elif normalizer_type == NormalizerType.STRIP:
        return StripNormalizer(strip_left=True, strip_right=True)
    elif normalizer_type == NormalizerType.REPLACE:
        return ReplaceNormalizer(pattern=StringPattern(String="a"), content="replacement")
    elif normalizer_type == NormalizerType.STRIPACCENTS:
        return StripAccentsNormalizer()
    elif normalizer_type == NormalizerType.PRECOMPILED:
        return PrecompiledNormalizer(precompiled_charsmap="precompiled_charsmap")
    else:
        raise ValueError(f"Unknown normalizer type: {normalizer_type}")


@pytest.mark.parametrize(
    "normalizer_type",
    [
        NormalizerType.BYTELEVEL,
        NormalizerType.BERTNORMALIZER,
        NormalizerType.LOWERCASE,
        NormalizerType.NFC,
        NormalizerType.NFD,
        NormalizerType.NFKC,
        NormalizerType.NFKD,
        NormalizerType.NMT,
        NormalizerType.PREPEND,
        NormalizerType.STRIP,
        NormalizerType.REPLACE,
    ],
)
def test_normalizer(small_tokenizer_json: dict[str, Any], normalizer_type: NormalizerType) -> None:
    """
    Test that the small tokenizer JSON can be loaded and contains the expected structure.

    This test checks that the tokenizer JSON has the correct keys and types for its fields.
    """
    normalizer = _get_default_normalizer(normalizer_type)
    normalizer_dict = normalizer.model_dump()
    small_tokenizer_json["normalizer"] = normalizer_dict
    tokenizer = TokenizerModel.model_validate(small_tokenizer_json)

    assert tokenizer.normalizer is not None
    assert tokenizer.normalizer.type == normalizer_type

    # Implicit test. If this fails, the model is incorrect.
    tokenizer.to_tokenizer()


@pytest.mark.parametrize(
    "normalizer,should_normalize",
    [
        [_get_default_normalizer(NormalizerType.BERTNORMALIZER), True],
        [BertNormalizer(clean_text=False, handle_chinese_chars=True, strip_accents=False, lowercase=False), False],
        [_get_default_normalizer(NormalizerType.LOWERCASE), True],
        [_get_default_normalizer(NormalizerType.STRIP), False],
        [NormalizerSequence(normalizers=[_get_default_normalizer(NormalizerType.LOWERCASE)]), True],
        [
            NormalizerSequence(
                normalizers=[NormalizerSequence(normalizers=[_get_default_normalizer(NormalizerType.LOWERCASE)])]
            ),
            True,
        ],
    ],
)
def test_lowercases(normalizer: Normalizer, should_normalize: bool) -> None:
    """Test whether the lowercases detection works."""
    assert normalizer.lowercases == should_normalize


@pytest.mark.parametrize(
    "normalizer,should_normalize",
    [
        [_get_default_normalizer(NormalizerType.BYTELEVEL), True],
        [_get_default_normalizer(NormalizerType.STRIP), False],
        [NormalizerSequence(normalizers=[_get_default_normalizer(NormalizerType.LOWERCASE)]), False],
        [
            NormalizerSequence(
                normalizers=[NormalizerSequence(normalizers=[_get_default_normalizer(NormalizerType.LOWERCASE)])]
            ),
            False,
        ],
    ],
)
def test_byte_normalizes(normalizer: Normalizer, should_normalize: bool) -> None:
    """Test whether the lowercases detection works."""
    assert normalizer.byte_normalizes == should_normalize



from __future__ import annotations

import logging
from enum import Enum
from typing import Annotated, Generic, Literal, TypeVar

from pydantic import BaseModel, Field

from skeletoken.merges import Merges
from skeletoken.vocabulary import UnigramVocabulary, Vocabulary

logger = logging.getLogger(__name__)


class ModelType(str, Enum):
    WORDPIECE = "WordPiece"
    BPE = "BPE"
    UNIGRAM = "Unigram"
    WORDLEVEL = "WordLevel"


VocabTypeVar = TypeVar("VocabTypeVar", Vocabulary, UnigramVocabulary)


class VocabMixinMethod(Generic[VocabTypeVar]):
    """Mixin to override token addition, removal etc."""

    vocab: VocabTypeVar

    def add_token(self, token: str, is_added_token: bool = False) -> None:
        """Add a token to the vocabulary."""
        self.vocab.add_token(token)

    def replace_token(self, old_token: str, new_token: str, is_added_token: bool = False) -> None:
        """Replace a token in the vocabulary."""
        self.vocab.replace_token(old_token, new_token)

    def remove_token(self, token: str) -> None:
        """Remove a token from the vocabulary."""
        self.vocab.remove_token(token)


class WordPiece(BaseModel, VocabMixinMethod[Vocabulary]):
    """Data model representing a WordPiece vocabulary."""

    type: Literal[ModelType.WORDPIECE] = ModelType.WORDPIECE
    vocab: Vocabulary
    unk_token: str
    continuing_subword_prefix: str
    max_input_chars_per_word: int = 100

    def to_greedy(self) -> WordPiece:
        """Convert the WordPiece model to a greedy version."""
        return self


class BPE(BaseModel, VocabMixinMethod[Vocabulary]):
    """Data model representing a BPE vocabulary."""

    type: Literal[ModelType.BPE] = ModelType.BPE
    merges: Merges
    vocab: Vocabulary
    dropout: float | None
    unk_token: str | None
    continuing_subword_prefix: str | None
    end_of_word_suffix: str | None
    fuse_unk: bool
    byte_fallback: bool
    ignore_merges: bool

    def to_greedy(self) -> WordPiece:
        """Convert the BPE model to a greedy WordPiece model."""
        if not self.unk_token:
            logger.warning("BPE model has no unk_token, using the first token in the vocab.")
            unk_token = self.vocab.sorted_vocabulary[0]
        else:
            unk_token = self.unk_token
        return WordPiece(
            vocab=self.vocab,
            unk_token=unk_token,
            continuing_subword_prefix=self.continuing_subword_prefix or "",
            max_input_chars_per_word=100,
        )

    def add_token(self, token: str, is_added_token: bool = False) -> None:
        """Add a token to the vocabulary."""
        self.vocab.add_token(token)
        if is_added_token:
            return
        self.merges._add_merges_for_token(token)
        new_tokens = sorted(self.merges._all_merge_tokens - set(self.vocab.vocabulary))
        for token in new_tokens:
            self.vocab.add_token(token)

    def replace_token(self, old_token: str, new_token: str, is_added_token: bool = False) -> None:
        """Replace a token in the vocabulary."""
        self.vocab.replace_token(old_token, new_token)
        # Added tokens do not require merge updates.
        if is_added_token:
            return
        new_tokens = self.merges._add_merges_for_token(new_token)
        for token in new_tokens:
            if token not in self.vocab.vocabulary:
                self.vocab.add_token(token)

    def remove_token(self, token: str) -> None:
        """Remove a token from the vocabulary."""
        self.vocab.remove_token(token)
        self.merges._remove_merges_for_token(token)


class Unigram(BaseModel, VocabMixinMethod[UnigramVocabulary]):
    """Data model representing a Unigram vocabulary."""

    type: Literal[ModelType.UNIGRAM] = ModelType.UNIGRAM
    vocab: UnigramVocabulary
    unk_id: int | None
    byte_fallback: bool

    def to_greedy(self) -> WordPiece:
        """Convert the Unigram model to a greedy WordPiece model."""
        if self.unk_id is None:
            unk_token = self.vocab.root[0][0]  # Use the first token as unk_token
            logger.warning("Unigram model has no `unk_id`, using the first token in the vocab.")
        else:
            unk_token = self.vocab.root[self.unk_id][0]
        return WordPiece(
            vocab=Vocabulary({token: idx for idx, (token, _) in enumerate(self.vocab.root)}),
            unk_token=unk_token,
            continuing_subword_prefix="",
            max_input_chars_per_word=100,
        )

    @property
    def unk_token(self) -> str | None:
        """Return the unknown token, if any."""
        if self.unk_id is None:
            return None
        return self.vocab.sorted_vocabulary[self.unk_id]

    @unk_token.setter
    def unk_token(self, token: str | None) -> None:
        """Set the unknown token."""
        if token is None:
            self.unk_id = None
        else:
            self.unk_id = self.vocab.vocabulary[token]


class WordLevel(BaseModel, VocabMixinMethod[Vocabulary]):
    """Data model representing a WordLevel vocabulary."""

    type: Literal[ModelType.WORDLEVEL] = ModelType.WORDLEVEL
    vocab: Vocabulary
    unk_token: str

    def to_greedy(self) -> WordPiece:
        """Convert the WordLevel model to a greedy WordPiece model."""
        return WordPiece(
            vocab=self.vocab,
            unk_token=self.unk_token,
            continuing_subword_prefix="",
            max_input_chars_per_word=100,
        )


Model = WordPiece | BPE | Unigram | WordLevel
ModelDiscriminator = Annotated[Model, Field(discriminator="type")]


def get_subword_prefix_token(model: Model) -> str | None:
    """Get the prefix token from the model, if any."""
    # Only WordPiece and BPE models have these.
    if isinstance(model, WordPiece):
        return model.continuing_subword_prefix
    elif isinstance(model, BPE):
        return model.continuing_subword_prefix
    return None


MODELS_THAT_NEED_UNK = (WordPiece, WordLevel)



import logging
from typing import Any

import pytest

from skeletoken.base import TokenizerModel
from skeletoken.postprocessors import (
    BertPostProcessor,
    ByteLevelPostProcessor,
    PostProcessor,
    PostProcessorSequence,
    PostProcessorType,
    RobertaPostProcessor,
    SpecialTokenInfo,
    TemplatePostProcessor,
    Token,
    TokenSequence,
    TokenType,
    get_bos_token_from_post_processor,
    get_eos_token_from_post_processor,
    maybe_replace_token_in_post_processor,
)


def _get_default_postprocessor(post_processor_type: PostProcessorType) -> PostProcessor:  # noqa: C901
    """Helper function to get the default instantiation of a normalizer."""
    if post_processor_type == PostProcessorType.BYTE_LEVEL:
        return ByteLevelPostProcessor(trim_offsets=True, add_prefix_space=False, use_regex=False)
    elif post_processor_type == PostProcessorType.BERT_PROCESSING:
        return BertPostProcessor(sep=("[SEP]", 1), cls=("[CLS]", 0))
    elif post_processor_type == PostProcessorType.ROBERTA_PROCESSING:
        return RobertaPostProcessor(sep=("[SEP]", 1), cls=("[CLS]", 0), trim_offsets=True, add_prefix_space=False)
    elif post_processor_type == PostProcessorType.TEMPLATE_PROCESSING:
        return TemplatePostProcessor(
            single=(
                Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
                Token(id="sequence", type_id=2, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=1, type=TokenType.SPECIAL),
            ),
            pair=(
                Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
                Token(id="sequence", type_id=2, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=1, type=TokenType.SPECIAL),
                Token(id="sequence2", type_id=2, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=1, type=TokenType.SPECIAL),
            ),
            special_tokens={
                "special_begin": SpecialTokenInfo(id="special_begin", ids=[0], tokens=["[BEGIN]"]),
                "special_end": SpecialTokenInfo(id="special_end", ids=[1], tokens=["[END]"]),
                "sequence": SpecialTokenInfo(id="sequence", ids=[2], tokens=["[SEQ]"]),
            },
        )
    elif post_processor_type == PostProcessorType.SEQUENCE:
        return PostProcessorSequence(
            processors=[ByteLevelPostProcessor(trim_offsets=True, add_prefix_space=False, use_regex=False)]
        )
    else:
        raise ValueError(f"Unknown normalizer type: {post_processor_type}")


@pytest.mark.parametrize("post_processor_type", [PostProcessorType.BYTE_LEVEL, PostProcessorType.BERT_PROCESSING])
def test_post_processor(small_tokenizer_json: dict[str, Any], post_processor_type: PostProcessorType) -> None:
    """
    Test that the small tokenizer JSON can be loaded and contains the expected structure.

    This test checks that the tokenizer JSON has the correct keys and types for its fields.
    """
    normalizer = _get_default_postprocessor(post_processor_type)
    normalizer_dict = normalizer.model_dump()
    small_tokenizer_json["post_processor"] = normalizer_dict
    tokenizer = TokenizerModel.model_validate(small_tokenizer_json)

    assert tokenizer.post_processor is not None
    assert tokenizer.post_processor.type == post_processor_type

    # Implicit test. If this fails, the model is incorrect.
    tokenizer.to_tokenizer()


def _get_no_eos_template() -> TemplatePostProcessor:
    """Gets a template processor without eos."""
    return TemplatePostProcessor(
        single=(
            Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
            Token(id="sequence", type_id=0, type=TokenType.SEQUENCE),
        ),
        pair=(
            Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
            Token(id="sequence1", type_id=0, type=TokenType.SEQUENCE),
            Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
            Token(id="sequence2", type_id=1, type=TokenType.SEQUENCE),
        ),
        special_tokens={
            "special_begin": SpecialTokenInfo(id="special_begin", ids=[0], tokens=["[BEGIN]"]),
        },
    )


def _get_no_bos_template() -> TemplatePostProcessor:
    """Gets a template processor without bos."""
    return TemplatePostProcessor(
        single=(
            Token(id="sequence", type_id=0, type=TokenType.SEQUENCE),
            Token(id="special_end", type_id=0, type=TokenType.SPECIAL),
        ),
        pair=(
            Token(id="sequence", type_id=0, type=TokenType.SEQUENCE),
            Token(id="special_end", type_id=0, type=TokenType.SPECIAL),
            Token(id="sequence2", type_id=1, type=TokenType.SEQUENCE),
            Token(id="special_end", type_id=0, type=TokenType.SPECIAL),
        ),
        special_tokens={
            "special_end": SpecialTokenInfo(id="special_end", ids=[1], tokens=["[END]"]),
        },
    )


@pytest.mark.parametrize(
    "post_processor,result",
    [
        (_get_default_postprocessor(PostProcessorType.SEQUENCE), None),
        (_get_default_postprocessor(PostProcessorType.BYTE_LEVEL), None),
        (_get_default_postprocessor(PostProcessorType.BERT_PROCESSING), ["[CLS]"]),
        (_get_default_postprocessor(PostProcessorType.ROBERTA_PROCESSING), ["[CLS]"]),
        (_get_default_postprocessor(PostProcessorType.TEMPLATE_PROCESSING), ["[BEGIN]"]),
        (_get_no_eos_template(), ["[BEGIN]"]),
        (_get_no_bos_template(), None),
    ],
)
def test_get_bos_token_from_post_processor(post_processor: PostProcessor, result: str | None) -> None:
    """Tests getting the bos token from the post processor."""
    bos_token = get_bos_token_from_post_processor(post_processor)
    assert bos_token == result


@pytest.mark.parametrize(
    "post_processor,result",
    [
        (_get_default_postprocessor(PostProcessorType.SEQUENCE), None),
        (_get_default_postprocessor(PostProcessorType.BYTE_LEVEL), None),
        (_get_default_postprocessor(PostProcessorType.BERT_PROCESSING), ["[SEP]"]),
        (_get_default_postprocessor(PostProcessorType.ROBERTA_PROCESSING), ["[SEP]"]),
        (_get_default_postprocessor(PostProcessorType.TEMPLATE_PROCESSING), ["[END]"]),
        (_get_no_eos_template(), None),
        (_get_no_bos_template(), ["[END]"]),
    ],
)
def test_get_eos_token_from_post_processor(post_processor: PostProcessor, result: str | None) -> None:
    """Tests getting the eos token from the post processor."""
    eos_token = get_eos_token_from_post_processor(post_processor)
    assert eos_token == result


@pytest.mark.parametrize(
    "post_processor,old_token,new_token",
    [
        (_get_default_postprocessor(PostProcessorType.SEQUENCE), "a", "b"),
        (_get_default_postprocessor(PostProcessorType.BYTE_LEVEL), "a", "b"),
        (_get_default_postprocessor(PostProcessorType.BERT_PROCESSING), "[CLS]", "[AB]"),
        (_get_default_postprocessor(PostProcessorType.ROBERTA_PROCESSING), "[CLS]", "[AB]"),
        (_get_default_postprocessor(PostProcessorType.BERT_PROCESSING), "[SEP]", "[AB]"),
        (_get_default_postprocessor(PostProcessorType.ROBERTA_PROCESSING), "[SEP]", "[AB]"),
        (_get_default_postprocessor(PostProcessorType.TEMPLATE_PROCESSING), "[END]", "[END]"),
    ],
)
def test_maybe_replace_token_in_post_processor(post_processor: PostProcessor, old_token: str, new_token: str) -> None:
    """Tests maybe replace the token in post processor."""
    result = maybe_replace_token_in_post_processor(old_token, new_token, 11, post_processor)
    if isinstance(result, PostProcessorSequence):
        # No change, because it has a single post-processor.
        assert result == post_processor
    if isinstance(result, (RobertaPostProcessor, BertPostProcessor)) and old_token == "[CLS]":
        assert result.cls == (new_token, 11)
        assert result.sep == ("[SEP]", 1)
    if isinstance(result, (RobertaPostProcessor, BertPostProcessor)) and old_token == "[SEP]":
        assert result.cls == ("[CLS]", 0)
        assert result.sep == (new_token, 11)
    if isinstance(result, TemplatePostProcessor):
        assert result.special_tokens["special_end"].tokens == [new_token]


def test_parse_token() -> None:
    """Test the Token parsing."""
    t = Token(id="A", type_id=0, type=TokenType.SPECIAL)
    assert t.id == "A"
    assert t.type_id == 0
    assert t.type == TokenType.SPECIAL

    t_dict = {"id": "B", "type_id": 1, "type": "Sequence"}
    t2 = Token.model_validate(t_dict)
    assert t2.id == "B"
    assert t2.type_id == 1
    assert t2.type == TokenType.SEQUENCE

    t_dict = {"Sequence": {"id": "B", "type_id": 1}}
    t2 = Token.model_validate(t_dict)
    assert t2.id == "B"
    assert t2.type_id == 1
    assert t2.type == TokenType.SEQUENCE

    t3 = Token.model_validate(t2)
    assert t3 == t2

    with pytest.raises(TypeError):
        Token.model_validate(11)  # type: ignore[arg-type]


def test_unequal_special_token(caplog) -> None:
    """Test that unequal special tokens are not equal."""
    with caplog.at_level(logging.WARNING):
        SpecialTokenInfo(id="A", ids=[0], tokens=["a", "b"])

    assert "ids and tokens must have the same length." in caplog.text
    assert caplog.records[0].levelname == "WARNING"


def test_template_incorrect_tokens() -> None:
    """Test that incorrect tokens in template raise errors."""
    with pytest.raises(ValueError):
        TemplatePostProcessor(
            single=(
                Token(id="missing_special", type_id=0, type=TokenType.SPECIAL),
                Token(id="sequence", type_id=2, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=1, type=TokenType.SPECIAL),
            ),
            pair=(
                Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
                Token(id="sequence", type_id=2, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=1, type=TokenType.SPECIAL),
            ),
            special_tokens={
                "special_begin": SpecialTokenInfo(id="special_begin", ids=[0], tokens=["[BEGIN]"]),
                "special_end": SpecialTokenInfo(id="special_end", ids=[1], tokens=["[END]"]),
            },
        )
    with pytest.raises(ValueError):
        TemplatePostProcessor(
            single=(Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),),
            pair=(Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),),
            special_tokens={
                "special_begin": SpecialTokenInfo(id="sapcco", ids=[0], tokens=["[BEGIN]"]),
                "special_end": SpecialTokenInfo(id="special_end", ids=[1], tokens=["[END]"]),
            },
        )


@pytest.mark.parametrize(
    "single,pair",
    [
        [(), ()],
        [
            (
                Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
                Token(id="sequence", type_id=0, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=0, type=TokenType.SPECIAL),
            ),
            (
                Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
                Token(id="sequence", type_id=0, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=0, type=TokenType.SPECIAL),
            ),
        ],
        [
            (
                Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
                Token(id="sequence", type_id=0, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=0, type=TokenType.SPECIAL),
            ),
            (
                Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
                Token(id="sequence1", type_id=0, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=0, type=TokenType.SPECIAL),
                Token(id="sequence1", type_id=0, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=0, type=TokenType.SPECIAL),
            ),
        ],
        [
            (
                Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
                Token(id="sequence", type_id=2, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=1, type=TokenType.SPECIAL),
            ),
            (
                Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
                Token(id="sequence1", type_id=2, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=1, type=TokenType.SPECIAL),
                Token(id="sequence2", type_id=1, type=TokenType.SEQUENCE),
                Token(id="special_begin", type_id=0, type=TokenType.SPECIAL),
                Token(id="sequence3", type_id=1, type=TokenType.SEQUENCE),
                Token(id="special_end", type_id=1, type=TokenType.SEQUENCE),
            ),
        ],
    ],
)
def test_template_creation_failure(single: TokenSequence, pair: TokenSequence) -> None:
    """Tests incorrect template creation."""
    with pytest.raises(ValueError):
        TemplatePostProcessor(
            single=single,
            pair=pair,
            special_tokens={
                "special_begin": SpecialTokenInfo(id="special_begin", ids=[0], tokens=["[BEGIN]"]),
                "special_end": SpecialTokenInfo(id="special_end", ids=[1], tokens=["[END]"]),
            },
        )



from enum import Enum

from pydantic import BaseModel


class TruncationDirection(str, Enum):
    LEFT = "Left"
    RIGHT = "Right"


class TruncationStrategy(str, Enum):
    LONGEST_FIRST = "LongestFirst"
    ONLY_FIRST = "OnlyFirst"
    ONLY_SECOND = "OnlySecond"


class Truncation(BaseModel):
    """
    Defines a truncation strategy to use.

    Attributes
    ----------
    direction : TruncationDirection
        The direction from which to truncate. This can be left or right.
    max_length : int
        The maximum length of the sequence after truncation.
    strategy : TruncationStrategy
        The strategy to use for truncation. OnlyFirst will only truncate from the
        first sequence, OnlySecond will only truncate from the second one, while
        LongestFirst will first truncate from the longest sequence, and then truncate
        the remainder from the shorter one.
    stride : int
        The stride to use. Stride here means how much overlap to leave in between
        sequential chunks. For example, if stride is 5, then each chunk contains 5
        chunks of the previous one at the start.

    """

    direction: TruncationDirection
    max_length: int
    strategy: TruncationStrategy
    stride: int



from pydantic import BaseModel, RootModel


class AddedToken(BaseModel):
    """
    Represents an added token in a tokenizer.

    An added token can be a special token or a regular token that is not necessarily part of the original vocabulary.
    Note that AddedToken can be used to represent multiword units. For example, a token like
    "New York" can be represented as a single AddedToken with content "New York".

    Multiword units are only supported if single_word is set to False.

    Attributes
    ----------
    content : str
        The string content of the token.
    single_word : bool
        Indicates if the token is a single word.
        If it is not a single word, it may be a subword or a character.
        e.g., if single_word is False, the token could be a subword like "ing".
    lstrip : bool
        If set, the token will be stripped from the left side.
    rstrip : bool
        If set, the token will be stripped from the right side.
    normalized : bool
        If set, the token form is searched _before_ normalization and pretokenization happens.
        For example, if you use a lowercase normalizer, the token "Hello" will be found as "hello",
        but only if normalized is set to True. If it is False, "Hello" will be found as is.
    special : bool
        If set, the token is a special token.
        Special tokens are skipped during decoding, and are represented as single tokens in the vocabulary.
        Special tokens are typically used to represent specific concepts or actions,
        such as [CLS] for classification or [SEP] for separation.
    id : int
        The unique identifier for the token.

    """

    content: str
    single_word: bool
    lstrip: bool
    rstrip: bool
    normalized: bool
    special: bool
    id: int


class AddedTokens(RootModel[list[AddedToken]]):
    """Represents a collection of AddedTokens."""

    def __getitem__(self, index: int) -> AddedToken:
        """Get an addedtoken by index."""
        return self.root[index]

    def get_token(self, token: str) -> AddedToken | None:
        """Returns the added token for a given form."""
        for added_token in self.root:
            if added_token.content == token:
                return added_token
        return None

    def maybe_add_token(
        self,
        token: str,
        id: int,
        is_special: bool = False,
        normalized: bool = False,
        single_word: bool = True,
        lstrip: bool = True,
        rstrip: bool = True,
    ) -> None:
        """Adds a new added token."""
        # If the token already exists, update it. Don't create a new one.
        if added_token := self.get_token(token):
            added_token.special = is_special
            added_token.normalized = normalized
            added_token.single_word = single_word
            added_token.lstrip = lstrip
            added_token.rstrip = rstrip
            added_token.id = id
        else:
            new_token = AddedToken(
                content=token,
                special=is_special,
                normalized=normalized,
                single_word=single_word,
                lstrip=lstrip,
                rstrip=rstrip,
                id=id,
            )
            self.root.append(new_token)

    def maybe_remove_token(self, token: str) -> None:
        """Removes the added token for a given form if it exists."""
        self.root = [added_token for added_token in self.root if added_token.content != token]

    def maybe_replace_token(self, old_token: str, new_token: str) -> None:
        """Replaces the added token for a given form, if it exists."""
        added_token = self.get_token(old_token)
        if added_token:
            added_token.content = new_token

    def get_special_tokens(self) -> list[AddedToken]:
        """Returns a list of all special added tokens."""
        return [token for token in self.root if token.special]

    def get_unnormalized_tokens(self) -> list[AddedToken]:
        """Returns a list of all unnormalized added tokens."""
        return [token for token in self.root if not token.normalized]

    def __len__(self) -> int:
        """Returns the number of added tokens."""
        return len(self.root)



from tokenizers.decoders import ByteLevel as ByteLevelDecoder
from tokenizers.normalizers import ByteLevel

from skeletoken.decase.byte_handlers import text_to_token_str, token_to_bytes


def test_token_to_bytes() -> None:
    """Test the conversion from token to bytes."""
    assert token_to_bytes("abba") == b"abba"
    assert token_to_bytes("Ġabba") == b" abba"
    assert token_to_bytes("ãģĵãĤĵãģ«ãģ¡ãģ¯") == b"\xe3\x81\x93\xe3\x82\x93\xe3\x81\xab\xe3\x81\xa1\xe3\x81\xaf"


def test_to_token_str() -> None:
    """Test the conversion from bytes to token string."""
    assert text_to_token_str("abba") == "abba"
    assert text_to_token_str(" abba") == "Ġabba"
    bytes = b"\xe3\x81\x93\xe3\x82\x93\xe3\x81\xab\xe3\x81\xa1\xe3\x81\xaf".decode("utf-8")
    assert text_to_token_str(bytes) == "ãģĵãĤĵãģ«ãģ¡ãģ¯"


def test_text_to_token_str() -> None:
    """Test the conversion from text to token string."""
    normalizer = ByteLevel()
    decoder = ByteLevelDecoder()
    normalized = normalizer.normalize_str("こんにちは")
    byte_form = token_to_bytes(normalized)
    assert byte_form == b"\xe3\x81\x93\xe3\x82\x93\xe3\x81\xab\xe3\x81\xa1\xe3\x81\xaf"
    assert text_to_token_str(byte_form.decode("utf-8")) == normalized

    assert decoder.decode([byte_form.decode("utf-8")]) == "こんにちは"
